import os
import click

from pymob.sim.config import Config
from pymob.utils import help

@click.command()
@click.option("-c", "--case_study", type=str, default="lotka_volterra_case_study", 
              help=help.case_study)
@click.option("-s", "--scenario", type=str, default="test_scenario", 
              help=help.scenario)
@click.option("-p", "--package", type=str, default="case_studies", 
              help=help.package)
@click.option("-l", "--logging", type=str, default=None)
@click.option("-f", "--logfile", type=str, default=None)
def main(case_study, scenario, package, logging, logfile):

    cfg = os.path.join(package, case_study, "scenarios", scenario, "settings.cfg")
    config = Config(cfg)
    config.case_study.name = case_study
    config.case_study.scenario = scenario
    config.import_casestudy_modules(reset_path=True)

    # update parameters from config file if they are specified
    if logging is not None: config.case_study.logging = logging
    if logfile is not None: config.case_study.logfile = logfile

    # import simulation      
    Simulation = config.import_simulation_from_case_study()
    sim = Simulation(config)
    sim.setup()

    # run the simulation
    evaluator = sim.dispatch(theta=sim.model_parameter_dict)
    evaluator()

    # store and process output
    sim.dump(results=evaluator.results)
    sim.plot(results=evaluator.results)

if __name__ == "__main__":
    main()


import pytest
import numpy as np
from scipy.stats import norm

from pymob.inference.scipy_backend import ScipyBackend
from pymob.sim.parameters import Param
from pymob.solvers.diffrax import JaxSolver

from tests.fixtures import init_lotka_volterra_case_study_hierarchical_from_script

def test_parameter_parsing_different_priors_on_species():
    sim = init_lotka_volterra_case_study_hierarchical_from_script()
    
    sim.config.model_parameters.alpha_species = Param(
        value=0.5, free=True, hyper=True,
        dims=('rabbit_species','experiment'),
        # take good care to specify hyperpriors correctly. 
        # Dimensions are broadcasted following the normal rules of 
        # numpy. The below means, in dimension one, we have two different
        # assumptions 1, and 3. Dimension one is the dimension of the rabbit species.
        # The specification loc=[1,3] would be understood as [[1,3]] and
        # be understood as the experiment dimension. Ideally, the dimensionality
        # is so low that you can be specific about the priors. I.e.:
        # scale = [[1,1,1],[3,3,3]]. This of course expects you know about
        # the dimensionality of the prior (i.e. the unique coordinates of the dimensions)
        prior="norm(loc=[[1],[3]],scale=0.1)" # type: ignore
    )
    # prey birth rate
    # to be clear, this says each replicate has a slightly varying birth
    # rate depending on the valley where it was observed. Seems legit.
    sim.config.model_parameters.alpha = Param(
        value=0.5, free=True, hyper=False,
        dims=('id',),
        prior="lognorm(s=0.1,scale=alpha_species[rabbit_species_index, experiment_index])" # type: ignore
    )

    # re initialize the observation with
    sim.define_observations_replicated_multi_experiment(n=120) # type: ignore
    sim.coordinates["time"] = np.arange(12)
    y0 = sim.parse_input("y0", drop_dims=["time"])
    sim.model_parameters["y0"] = y0

    inferer = ScipyBackend(simulation=sim)


    theta = inferer.sample_distribution()

    alpha_samples_cottontail = theta["alpha"][sim.observations["rabbit_species"] == "Cottontail"]
    alpha_samples_jackrabbit = theta["alpha"][sim.observations["rabbit_species"] == "Jackrabbit"]

    alpha_cottontail = np.mean(alpha_samples_cottontail)
    alpha_jackrabbit = np.mean(alpha_samples_jackrabbit)
    
    # test if the priors that were broadcasted to the replicates 
    # match the hyperpriors
    np.testing.assert_array_almost_equal(
        [alpha_cottontail, alpha_jackrabbit], [1, 3], decimal=1
    )

    sim.solver = JaxSolver
    sim.model_parameters["parameters"] = sim.config.model_parameters.value_dict
    sim.dispatch_constructor()
    e = sim.dispatch(theta=theta)
    e()

    # res_species = e.results.where(e.results.rabbit_species=="Cottontail", drop=True)
    # store simulated results

    # TODO: mark datavars as observed automatically
    # set observations and mark as observed. This 
    # could be automated by using length of data var > 0 to
    # trigger marking data vars as observed.
    rng = np.random.default_rng(1)
    
    
    obs = e.results
    obs.rabbits.values = rng.poisson(e.results.rabbits+1e-6)
    obs.wolves.values = rng.poisson(e.results.wolves+1e-6)
    obs.to_netcdf(
        f"{sim.data_path}/simulated_data_hierarchical_species_year.nc"
    )

    sim.observations = obs
    sim.config.data_structure.rabbits.observed = True
    sim.config.data_structure.wolves.observed = True

    # this is the conventional way to define error models. We center a lognormal
    # error model around the means of the distribution. The problem is: 
    # due to the large scale differences in rabbits and wolves, the log-likelihoods
    # end up very differently. Here the wolves data variable will basically
    # be meaningless, because the rabbits data variable is at such a high scale
    # Scaling alone also does not resolve this problem, because due to the dynamic
    # of the data variables, larger values will have a higher weight. This is not
    # right.
    sim.config.error_model.rabbits = "lognorm(scale=rabbits+EPS, s=0.1)"
    sim.config.error_model.wolves = "lognorm(scale=wolves+EPS, s=0.1)"
    sim.dispatch_constructor()
    # TODO: At this point, I should get a joint likelihood function and try
    # to fit the model with scipy minimize.


def test_parameter_parsing_different_priors_on_year():
    sim = init_lotka_volterra_case_study_hierarchical_from_script()
    
    sim.config.model_parameters.alpha_species = Param(
        value=0.5, free=True, hyper=True, dims=('rabbit_species','experiment'),
        prior="norm(loc=[[1,2,3]],scale=0.1)" # type: ignore
    )
    # prey birth rate
    sim.config.model_parameters.alpha = Param(
        value=0.5, free=True, hyper=True, dims=('id',),
        prior="lognorm(s=0.1,scale=alpha_species[rabbit_species_index, experiment_index])" # type: ignore
    )

    # re initialize the observation with
    sim.define_observations_replicated_multi_experiment(n=120) # type: ignore
    y0 = sim.parse_input("y0", drop_dims=["time"])
    sim.model_parameters["y0"] = y0

    inferer = ScipyBackend(simulation=sim)


    theta = inferer.sample_distribution()

    alpha_samples_2010 = theta["alpha"][sim.observations["experiment"] == "2010"]
    alpha_samples_2011 = theta["alpha"][sim.observations["experiment"] == "2011"]
    alpha_samples_2012 = theta["alpha"][sim.observations["experiment"] == "2012"]

    alpha_2010 = np.mean(alpha_samples_2010)
    alpha_2011 = np.mean(alpha_samples_2011)
    alpha_2012 = np.mean(alpha_samples_2012)
    
    # test if the priors that were broadcasted to the replicates 
    # match the hyperpriors
    np.testing.assert_array_almost_equal(
        [alpha_2010, alpha_2011, alpha_2012], [1, 2, 3], decimal=1
    )




if __name__ == "__main__":
    pass


import pytest
from click.testing import CliRunner
import tempfile
from pymob.simulation import SimulationBase, Config
from pymob.sim.config import DataVariable, Datastructure, configure
from pymob.sim.parameters import Param, RandomVariable, Expression
from pymob.utils.store_file import import_package
from pymob.solvers.scipy import solve_ivp_1d
from sympy import Function
import xarray as xr
import numpy as np
import os

scenario = "case_studies/lotka_volterra_case_study/scenarios/test_scenario_scripting_api"

def test_simulation():
    sim = SimulationBase()
    sim.config.case_study.name = "lotka_volterra_case_study"
    sim.config.case_study.scenario = "test_scenario_scripting_api"
    sim.config.case_study.observations = "simulated_data.nc"
    sim.config.case_study.data_path = None

    # load data before specifying dims
    sim.config.case_study.data = os.path.abspath("case_studies/lotka_volterra_case_study/data")
    sim.observations = xr.load_dataset(sim.config.input_file_paths[0])    

    # try wrong specification
    threw_error = None
    try:
        sim.config.data_structure.rabbits = DataVariable(dimensions=["hash"])
        sim.config.data_structure.wolves = DataVariable(dimensions=["time"])
        sim.observations = xr.load_dataset(sim.config.input_file_paths[0])
        threw_error = False
    except KeyError:
        threw_error = True

    assert threw_error

    sim.validate()

    sim.config.data_structure.rabbits = DataVariable(dimensions=["time"])
    sim.config.data_structure.wolves = DataVariable(dimensions=["time"])
    
    # load data by providing an absolute path
    sim.config.case_study.data = os.path.abspath("case_studies/lotka_volterra_case_study/data")
    sim.observations = xr.load_dataset(sim.config.input_file_paths[0])    

    # load data by providing a relative path
    sim.config.case_study.data = "case_studies/lotka_volterra_case_study/data"
    sim.observations = xr.load_dataset(sim.config.input_file_paths[0])    
    
    # load data by providing no path (the default 'data' directory in the case study)
    sim.config.case_study.data = None
    sim.observations = xr.load_dataset(sim.config.input_file_paths[0])    

    sim.config.case_study.output = None

    # resetting the path is important in case other case studies were imported
    # in the same testing session
    sim.config.import_casestudy_modules(reset_path=True)
    sim.config.simulation.model = "lotka_volterra"

    sim.solver = solve_ivp_1d
    sim.setup()
    sim.config.save(
        fp=f"{scenario}/test_settings.cfg",
        force=True, 
    )
    sim.save_observations(filename=sim.config.case_study.observations, force=True)

def test_load_generated_settings():
    sim = SimulationBase(f"{scenario}/test_settings.cfg")
    assert sim.config.case_study.name == "lotka_volterra_case_study"
    assert sim.config.case_study.scenario == "test_scenario_scripting_api"
    assert sim.config.case_study.package == "case_studies"
    assert sim.config.case_study.data == None
    assert sim.config.case_study.data_path == "case_studies/lotka_volterra_case_study/data"
    assert sim.config.case_study.output == None
    assert sim.config.case_study.output_path == \
        "case_studies/lotka_volterra_case_study/results/test_scenario_scripting_api"

def test_load_interpolated_settings():
    sim = SimulationBase(f"{scenario}/interp_settings.cfg")
    expected_output = \
        "./case_studies/lotka_volterra_case_study/results/test_scenario_scripting_api"
    assert sim.config.case_study.output == expected_output



def test_standalone_casestudy():
    wd = os.getcwd()
    case_study_name = "lotka_volterra_case_study_standalone"
    root = os.path.join(str(tempfile.tempdir), case_study_name)
    os.mkdir(root)
    os.chdir(root)
    
    # this is the syntax for setting up a standalone case study
    # currently root cannot be set with the config backend, but needs
    # to be specified with `chdir`
    sim = SimulationBase()
    sim.config.case_study.name = "."
    sim.config.case_study.scenario = "test_scenario_standalone"
    sim.config.case_study.package = "."

    os.makedirs(sim.config.case_study.output_path)
    os.makedirs(sim.config.case_study.data_path)
    os.makedirs(sim.config.case_study.scenario_path)
    sim.config.save(force=True)

    # test if all files exist and remove test directory
    os.chdir(wd)
    file_structure = [
        f"{tempfile.tempdir}/lotka_volterra_case_study_standalone",
        f"{tempfile.tempdir}/lotka_volterra_case_study_standalone/data",
        f"{tempfile.tempdir}/lotka_volterra_case_study_standalone/results",
        f"{tempfile.tempdir}/lotka_volterra_case_study_standalone/results/test_scenario_standalone",
        f"{tempfile.tempdir}/lotka_volterra_case_study_standalone/scenarios",
        f"{tempfile.tempdir}/lotka_volterra_case_study_standalone/scenarios/test_scenario_standalone",
        f"{tempfile.tempdir}/lotka_volterra_case_study_standalone/scenarios/test_scenario_standalone/settings.cfg",
    ]
    
    for p in reversed(file_structure):
        assert os.path.exists(p)
        if os.path.isdir(p):
            os.rmdir(p)
        else:
            os.remove(p)

def test_parameter_parsing():
    config = Config()

    io = "value=1.0 dims=[] min=0.0 max=3.0 hyper=False free=True"

    # test scripting input
    test = Param(value=1.0, min=0.0, max=3.0)
    config.model_parameters.test = test

    # test dict input
    config.model_parameters.test = test.model_dump(exclude_none=True)
    assert config.model_parameters.test == test # type: ignore

    # test validation
    config.model_parameters.test = io
    assert config.model_parameters.test == test # type: ignore

    # test serialization
    serialized = config.model_parameters.model_dump(mode="json")
    assert serialized == {"test": io}


def test_parameter_array():
    config = Config()

    io = "value=[1.0,2.0,3.0] dims=['test_dim'] hyper=False free=True"

    # test scripting input
    test = Param(value=np.array([1.0,2.0,3.0]))
    config.model_parameters.test = test

    # test dict input
    config.model_parameters.test = test.model_dump(exclude_none=True)
    assert config.model_parameters.test == test # type: ignore

    # test config file input
    config.model_parameters.test = io
    assert config.model_parameters.test == test  # type: ignore
    
    # test serialization
    serialized = config.model_parameters.model_dump(mode="json")
    assert serialized == {"test": io}

def test_prior():
    config = Config()

    io = "lognorm(scale=[1.0,1.0,a],s=1.5)"

    test_prior = RandomVariable(
        distribution="lognorm", 
        parameters={"scale": Expression("[1.0,1.0,a]"), "s": Expression("1.5")},
    )

    # test scripting input
    test_param = Param(value=np.array([1.0,2.0,3.0]))
    test_param.prior = test_prior

    # test dict input
    test_param.prior = test_prior.model_dump(exclude_none=True)
    assert test_param.prior == test_prior # type: ignore

    # test config file input
    test_param.prior = io
    assert test_param.prior == test_prior  # type: ignore
    
    # test serialization
    config.model_parameters.test = test_param
    serialized = test_prior.model_dump(mode="json")
    config.model_parameters.model_dump(mode="json")
    assert serialized == io

def test_parameter_array_with_prior():
    config = Config()

    io = "value=[1.0,2.0,3.0] dims=[] prior=lognorm(scale=[1.0,1.0,1.0],s=1.0) hyper=False free=True"
    test = Param(value=[1.0,2.0,3.0], prior="lognorm(scale=[1.0,1.0,1.0],s=1.0)") # type:ignore

    # test config file input
    config.model_parameters.test = io
    assert config.model_parameters.test == test  # type: ignore
    
    # test scripting input
    config.model_parameters.test = test

    # test dict input
    config.model_parameters.test = test.model_dump(exclude_none=True)
    assert config.model_parameters.test == test # type: ignore

    # test serialization
    serialized = config.model_parameters.model_dump(mode="json")
    assert serialized == {"test": io}



def test_model_parameters():
    config = Config()

    a = Param(value=1)
    b = Param(value=5, free=False)

    config.model_parameters.a = a
    config.model_parameters.b = b

    frmp = config.model_parameters.free
    fimp = config.model_parameters.fixed
    almp = config.model_parameters.all

    assert frmp == {"a": a}
    assert fimp == {"b": b}
    assert almp == {"a": a, "b":b}

def test_error_model():
    config = Config()

    io = "norm(loc=1,scale=2)"
    
    # test config file input
    test = RandomVariable(
        distribution="norm",
        parameters=dict(loc=Expression("1"),scale=Expression("2")),  # type:ignore
    )

    # test config file input
    config.error_model.test = io
    assert config.error_model.test == test  # type: ignore
    
    # test scripting input
    config.error_model.test = test

    # test dict input
    config.error_model.test = test.model_dump(exclude_none=True)
    assert config.error_model.test == test # type: ignore

    # test serialization
    serialized = config.error_model.model_dump(mode="json", exclude_none=True)
    assert serialized == {"test": io}

def test_error_model_with_obs():
    config = Config()

    io = "lognorm(scale=[1.0,1.0,1.0],s=1.0,obs=b/jnp.sqrt(2))"
    test = RandomVariable(
        distribution="lognorm",
        parameters=dict(scale=Expression("[1.0,1.0,1.0]"),s=Expression("1.0")),  # type:ignore
        obs=Expression("b/jnp.sqrt(2)")
    )

    # test config file input
    config.error_model.test = io
    assert config.error_model.test == test  # type: ignore
    
    # test scripting input
    config.error_model.test = test

    # test dict input
    config.error_model.test = test.model_dump(exclude_none=True)
    assert config.error_model.test == test # type: ignore

    # test serialization
    serialized = config.error_model.model_dump(mode="json")
    assert serialized == {"test": io}



def test_data_variables():
    config = Config()
    config.case_study.name = "lotka_volterra_case_study"
    config.case_study.scenario = "test_scenario_scripting_api"
    config.data_structure.wolves = DataVariable(dimensions=["time"], min=0)
    assert config.data_structure.data_variables == ["wolves"]
    config.save(force=True)
    
    config.data_structure = {"wolves": dict(dimensions = ["time"])} # type: ignore
    assert config.data_structure.data_variables == ["wolves"]


    config.data_structure.B = DataVariable(dimensions=["a", "b"], dimensions_evaluator=["b","a"])
    assert config.data_structure.dimdict == {"wolves": ["time"], "B": ["a", "b"]}
    assert config.data_structure.var_dim_mapper == {"wolves": [0], "B": [1,0]}



def test_commandline_api_configure():
    runner = CliRunner()
    
    args = [
        "--file=case_studies/lotka_volterra_case_study/scenarios/test_scenario_scripting_api/test_settings.cfg",
        "-o inference.numpyro.kernel=nuts",
        "-o simulation.y0 = wolves=wolves rabbits=rabbits",
        "-o inference.n_predictions=10",
        "-o model-parameters.test_parameter=",
    ]
    result = runner.invoke(configure, args)

    if result.exception is not None:
        raise result.exception


if __name__ == "__main__":
    pass


import numpy
import numpy as np
from numpy.typing import ArrayLike
from types import ModuleType
import xarray as xr
from typing import (
    Callable, Dict, List, Optional, Sequence, Literal, Tuple, Union
)
from frozendict import frozendict
from dataclasses import dataclass, field
import inspect
from scipy.ndimage import gaussian_filter1d
from diffrax import rectilinear_interpolation

@dataclass(frozen=True)
class SolverBase:
    """
    The idea of creating a solver as a class is that it is easier
    to pass on important arguments of the simulation relevant to the 
    Solver. Therefore a solver can access all attributes of an Evaluator
    """
    model: Callable
    dimensions: Tuple
    dimension_sizes: frozendict[str, int]
    parameter_dims: frozendict[str, Tuple[str, ...]]
    n_ode_states: int
    coordinates: frozendict[str, Tuple] = field(repr=False)
    coordinates_input_vars: frozendict[str, frozendict[str, frozendict[str, Tuple[Union[float,int,str], ...]]]]
    dims_input_vars: frozendict[str, frozendict[str, Tuple[str, ...]]]
    coordinates_indices: frozendict[str, tuple]
    data_variables: Tuple
    data_structure_and_dimensionality: frozendict[str, frozendict[str, int]]
    is_stochastic: bool
    post_processing: Callable
    solver_kwargs: frozendict = frozendict()
    indices: frozendict[str, Tuple] = field(repr=False, default=frozendict())

    x_dim: str = "time"
    batch_dimension: str = "batch_id"
    exclude_kwargs_model: Tuple[str, ...] = ("t", "x_in", "y", "X")
    exclude_kwargs_postprocessing: Tuple[str, ...] = ("t", "time", "interpolation", "results")

    # fields that are computed post_init
    x: Tuple[float] = field(init=False, repr=False)
    shapes_coordinates: Dict[str, int] = field(init=False, repr=False)
    shapes_parameter_coordinates: Dict[str, Tuple[int, ...]] = field(init=False, repr=False)
    shapes_input_vars_coordinates: frozendict[str, frozendict[str, Tuple[int, ...]]] = field(init=False, repr=False)
    len_batch_coordinate: int = field(init=False)
    x_shape_batched: Tuple[int, ...] = field(init=False)

    def __post_init__(self, *args, **kwargs):
        x = self.coordinates[self.x_dim]
        if not np.all(x[:-1] <= x[1:]):
            raise ValueError(
                f"x_dim '{self.x_dim}' must be sorted in ascending order."
            )
        object.__setattr__(self, "x", x)

        coord_dims = self._get_coordinate_unique_shapes()
        object.__setattr__(self, "shapes_coordinates", coord_dims)
        
        parameter_shapes = self._get_parameter_shapes()
        object.__setattr__(self, "shapes_parameter_coordinates", parameter_shapes)

        input_vars_shapes = self._get_input_vars_shapes()
        object.__setattr__(self, "shapes_input_vars_coordinates", input_vars_shapes)

        batch_length = self._get_batch_length()
        object.__setattr__(self, "len_batch_coordinate", batch_length)

        x_shape_batched = self._get_x_shape_batched()
        object.__setattr__(self, "x_shape_batched", x_shape_batched)

        # set extra attributes from solver_kwargs, which are specified through
        # the dispatch_constructor. Those don't receive post-processing

        self.test_matching_batch_dims()
        self.test_x_coordinates()
        

    def __call__(self, **kwargs):
        return self.solve(**kwargs)
    
    def _get_batch_length(self) -> int:
        batch_coordinates = self.coordinates.get(self.batch_dimension, [0])
        return len(batch_coordinates)

    def _get_x_shape_batched(self) -> Tuple[int, ...]:
        n_batch = self.len_batch_coordinate
        n_x = self.dimension_sizes[self.x_dim]
        return (n_batch, n_x)

    def _get_coordinate_unique_shapes(self, ) -> frozendict[str, int]:
        coordinate_shape_dict = {}
        for key, coords in self.coordinates.items():
            n_coords = len(coords)
            coordinate_shape_dict.update({key: n_coords})

        for key, coords in self.coordinates_indices.items():
            unique_coords = set(coords)
            n_coords = len(unique_coords)
            coordinate_shape_dict.update({key: n_coords})

        return frozendict(coordinate_shape_dict)

    def _get_parameter_shapes(self, ) -> frozendict[str, Tuple[int, ...]]:
        par_shape_dict = {}
        for par_name, par_dims in self.parameter_dims.items():
            dim_shape = []
            for d in par_dims:
                if d == self.batch_dimension and d not in self.dimension_sizes:
                    dim_size = 1
                else:
                    try:
                        dim_size = self.dimension_sizes[d]
                    except KeyError as err:
                        raise KeyError(
                            f"KeyError: Dimension '{d}' could not be found in "+
                            f"any dimension specified for the simulation "+
                            f"{self.dimension_sizes}. It must be either "+
                            "specified via the `sim.indices` dict."
                        )
                
                dim_shape.append(dim_size)
            par_shape_dict.update({par_name: tuple(dim_shape)})

        return frozendict(par_shape_dict)

    def _get_input_vars_shapes(self, ) -> frozendict[str, frozendict[str, Tuple[int, ...]]]:
        input_vars_shape_dict = {}
        for input_var_name, input_var_dict in self.dims_input_vars.items():
            data_var_shape_dict = {}
            for data_var_name, data_var_dims in input_var_dict.items():
                dim_shape = []
                for d in data_var_dims:
                    # look for the dimension in the coodinates of the input vars
                    coords_data_var = self.coordinates_input_vars[input_var_name][data_var_name]
                    coords = coords_data_var.get(d, None)
                    
                    # if it is found use its length as a dimension size
                    if coords is not None:
                        dim_size = len(coords)
                    
                    # if it is not found, see if the dim is the batch dimensions
                    # and handle the batch dimension size
                    # this is done, in cases where there is a batch dimension, 
                    # but it was not specified for the input, because it is 
                    # expected that it simply broadcasts to the batch dimension
                    elif d == self.batch_dimension:
                        if d not in self.dimension_sizes:
                            dim_size = 1
                        else:
                            dim_size = self.dimension_sizes[d]

                    # if the dimension is not in the coordinates of the input vars
                    # or not as a batch dimension in the dimension sizes dict
                    # something is severely wrong
                    else:
                        raise KeyError(
                            f"KeyError: Dimension: '{d}' for data_var: "+
                            f"'{data_var_name}' in input var: '{input_var_name}' "+
                            f"could not be found in the coordinate dimensions: "+
                            f"{list(coords_data_var.keys())}. "+
                            f"Make sure `sim.model_parameters['{input_var_name}']` "+
                            "is correctly specified. It is recommended to "+
                            "specify the model_input with `sim.config.simulation."+
                            f"{input_var_name} = ['...', '{data_var_name}=...', '...']` "+
                            f"and use `sim.parse_input(input='{input_var_name}', "+
                            f"reference_data=..., drop_dims=[...])` to parse the "+
                            "input."
                        )
                    
                    dim_shape.append(dim_size)
                data_var_shape_dict.update({data_var_name: tuple(dim_shape)})
            input_vars_shape_dict.update({
                input_var_name: frozendict(data_var_shape_dict)
            })
        return frozendict(input_vars_shape_dict)

    # def _get_input_vars_shapes(self, ) -> frozendict[str, frozendict[str, Tuple[int, ...]]]:
    #     frozen_shapes_input_vars = frozendict({
    #         k_input_var: frozendict({
    #             k_datavar: tuple([
    #                 len(v_coord)
    #                 for _, v_coord in v_datavar.items()
    #             ]) 
    #             for k_datavar, v_datavar in v_input_var.items()
    #         }) 
    #         for k_input_var, v_input_var in self.coordinates_input_vars.items()
    #     })

    #     return frozen_shapes_input_vars

    #     raise NotImplementedError(
    #         "Currently, it is expected that x_in is already in the correct "+
    #         "shape (minus the batch dimension), when passed to the Evaluator. "+
    #         "This is why no definite shapes can be defined on initialization. "+
    #         "The only broadcasting that can be sensibly made is to promote the "+
    #         "input vectors to the batch dimension, if this has not yet been done."
    #     )

    def test_matching_batch_dims(self):
        bc = self.coordinates.get(self.batch_dimension, None)

        if bc is not None:
            matching_batch_coords_if_present = [
                v[self.batch_dimension] == bc 
                for k, v in self.coordinates_input_vars.items() 
                if self.batch_dimension in v
            ]

            if not all(matching_batch_coords_if_present):
                raise IndexError(
                    f"Batch coordinates '{self.batch_dimension}' of input "+
                    "variables do not have the same size "+
                    "as the batch dimension of the observations."
                )

    def test_x_coordinates(self):
        x = self.coordinates[self.x_dim]
        if len(self.coordinates_input_vars["x_in"]) == 0:
            return
        
        x_xin = [
            np.max(v.get(self.x_dim, [0])) 
            for v in self.coordinates_input_vars["x_in"].values()
        ]

        if np.max(x) > np.max(x_xin):
            raise AssertionError(
                f"The {self.x_dim}-coordinate on the observations (sim.coordinates) "+
                f"goes to a higher {self.x_dim} than the {self.x_dim}-coordinate "+
                "of the model_parameters['x_in']. "+
                "Make sure to run the simulation only until the provided x_in "+
                "values, or extend the x_in values until the required time"
            )

    def preprocess_parameters(self, parameters, num_backend: ModuleType=numpy):
        ode_args = mappar(
            self.model, 
            parameters, 
            exclude=self.exclude_kwargs_model, 
            to="dict"
        )
        ode_args_broadcasted = self._broadcast_args(
            arg_dict=frozendict(ode_args), # type: ignore
            num_backend=num_backend
        )
        
        pp_args = mappar(
            self.post_processing, 
            parameters, 
            exclude=self.exclude_kwargs_postprocessing, 
            to="dict"
        )
        pp_args_broadcasted = self._broadcast_args(
            arg_dict=frozendict(pp_args), # type: ignore
            num_backend=num_backend
        )

        return ode_args_broadcasted, pp_args_broadcasted


    def _broadcast_args(self, arg_dict: frozendict[str, numpy.ndarray], num_backend: ModuleType=numpy):
        # simply broadcast the parameters along the batch dimension
        # if there is no other index provided
        n_batch = self.len_batch_coordinate

        args = []
        for arg_name, arg in arg_dict.items():
            # you can expect that any of the expected parameter_shapes have the
            # size of the batch_dimension in the first axis.
            # Note that this is the taget shape without any extended dimension.
            # Dimensions are only extended, if there is no batch dimension already
            # present
            target_shape = self.shapes_parameter_coordinates.get(arg_name, (n_batch, ))

            # make sure the argument is an array with one dimension
            # promoting to two dimension (including a batch dimension will be done)
            # in the different conditional branches
            arg_promoted = num_backend.array(arg, ndmin=1, dtype=float)
            
            # if the size of the 1st dimension of the argument array 
            # is identical to the size of the specified first dimension
            # we know that the input array was constructed to match the size.
            # The length of the batch coordinates is at least 1
            # There are weird cases in which the input arrays are square,
            # and the dimensions of the parameter have been specified in a 
            # different order (other than batch_dimension first)
            # This problem has been fixed, as the dimensional order is now
            # checked in the Evaluator.__init__
            if arg_promoted.shape[0] == target_shape[0]:
                # if the dimensionality of the argument array is 1
                # we add a dummy dimension at the end, in order
                # to harmonize it with arguments that are vectors
                if arg_promoted.ndim == 1:
                    arg_broadcasted = num_backend.expand_dims(arg_promoted, -1)
                else:
                    # if greater zero (zero dim not possible because of the 
                    # promotion to 1D arrays at the beginning of the loop)
                    # then it is assumed that the array correctly contains 
                    # more than one value for each id in the batch dimension
                    # i.e. vector, matrix or nd-array parameters in the ODE
                    # We leave the array as is
                    arg_broadcasted = arg_promoted

            elif (
                # this is when passed and expected shapes have the same number
                # of dimensions
                arg_promoted.shape[0] == 1 
                # this is when the shape of the later dimensions match.
                # I.e. the batch_dimension is not in the passed argument values
                # This happens frequently, when parameters are not broadcasted
                # to the batch dimension before being passed to the solver
                #
                # An example for the syntax below:
                # arg_promoted.shape >>> (2,5) 
                # target_shape >>> (10,2,5)
                # len(arg_promoted.shape) >>> 2
                # target_shape[-2:] >>> (2,5)
                # (2,5) == (2,5) >>> True
                or arg_promoted.shape == target_shape[-len(arg_promoted.shape):]
            ):
                # Note that this will also broadcast parameters across multiple
                # dimension e.g. (5,) -> (10,2,5)
                if arg_promoted.ndim > len(self.shapes_parameter_coordinates[arg_name]):
                    raise ValueError(
                        f"Parameter '{arg_name}' values have shape "+
                        f"{arg_promoted.shape}. This is in conflict with the "+
                        f"specified shape {self.shapes_parameter_coordinates[arg_name]} "+
                        f"from the dimensions {self.parameter_dims[arg_name]}."
                    )
                # the operation will broadcase argument arrays that have the
                # same number of dimensions (and the first dimension is )
                arg_broadcasted = num_backend.broadcast_to(
                    arg_promoted, shape=self.shapes_parameter_coordinates[arg_name]
                )

                # also here we apply broadcasting to the array if the result has 
                # only one dimension
                if arg_broadcasted.ndim == 1:
                    arg_broadcasted = num_backend.expand_dims(arg_broadcasted, -1)
            else:
                raise ValueError(
                    f"The values of parameter '{arg_name}' with the shape "+
                    f"{arg_promoted.shape} could not be broadcasted to the "+
                    f"specified shape {self.shapes_parameter_coordinates[arg_name]} "+
                    f"from the dimensions {self.parameter_dims[arg_name]}. "+
                    f"Make sure you add the missing dimensions in the "+
                    f"parameter specification Param(..., dims=(...,)) "+
                    f"and handle parameter coordinates appropriately. "+
                    "Parameter dimension coordinates can be specified either "+
                    f"1) Add the missing dimension in sim.config.parameters.{arg_name} "+
                    "and sim.coordinates['MISSING_DIM'] = [...]"+
                    "2) Add the missing dimension in "+
                    "sim.config.data_structure.ANY_DATA_VAR and sim.coordinates['MISSING_DIM'] "+
                    "3) dimension in sim.indices "+
                    f"the batch dimension {self.batch_dimension} is added "+
                    f"automatically."
                )
            
            args.append(arg_broadcasted)

        return tuple(args)


    def preprocess_x_in(self, x_in, num_backend:ModuleType=numpy):
        X_in_list = []
        for x_in_var, x_in_vals in x_in.items():
            # parse to array
            x_in_x = num_backend.array(
                self.coordinates_input_vars["x_in"][x_in_var][self.x_dim], 
                dtype=float,
                ndmin=1
            )
            x_in_y = num_backend.array(x_in_vals, ndmin=1, dtype=float)

            # broadcast x
            x_in_batched_shape = (self.len_batch_coordinate, *x_in_x.shape)
            x_in_x_broadcasted = num_backend.broadcast_to(x_in_x, x_in_batched_shape)

            # broadcast y
            y_in_batched_shape = self.shapes_input_vars_coordinates["x_in"][x_in_var]
            x_in_y_broadcasted = num_backend.broadcast_to(x_in_y, y_in_batched_shape) 

            # also here we apply broadcasting to the array if the result has 
            # only one dimension
            if x_in_y_broadcasted.ndim == 1:
                x_in_y_broadcasted = num_backend.expand_dims(
                    x_in_y_broadcasted, -1
                )

            # combine xs and ys to make them ready for interpolation
            X_in = [
                num_backend.array(v) for v in 
                [x_in_x_broadcasted, x_in_y_broadcasted]
            ]

            X_in_list.append(X_in)

        return X_in_list
    
    def preprocess_y_0(self, y0, num_backend:ModuleType=numpy):
        Y0 = []

        for y0_var, y0_vals in y0.items():
            y0_vals_promoted = num_backend.array(y0_vals, ndmin=1, dtype=float)
            
            # wrap y0 data in a dummy batch dim if the batch dim is not
            # included in the coordinates
            y0_var_shape = self.shapes_input_vars_coordinates["y0"][y0_var]
            y0_vals_broadcasted = num_backend.broadcast_to(
                y0_vals_promoted, 
                y0_var_shape
            ) 

            # also here we apply broadcasting to the array if the result has 
            # only one dimension
            if y0_vals_broadcasted.ndim == 1:
                y0_vals_broadcasted = num_backend.expand_dims(
                    y0_vals_broadcasted, -1
                )

            Y0.append(y0_vals_broadcasted)
        return Y0


    def solve(self):
        raise NotImplementedError("Solver must implement a solve method.")

def mappar(
    func, 
    parameters: Dict[str,float|int|List|Tuple], 
    exclude=[], 
    to:Literal["tuple","dict","names"]="tuple"
) -> Tuple|Dict:
    func_signature = inspect.signature(func).parameters.keys()
    model_param_signature = [p for p in func_signature if p not in exclude]
    if to == "tuple":
        model_args = [parameters.get(k) for k in model_param_signature]
        return tuple(model_args)
    elif to == "dict":
        return {k: parameters.get(k) for k in model_param_signature}
    elif to == "names":
        return tuple(model_param_signature)

    raise NotImplementedError(f"'to={to}' is not implemented for 'mappar'")


def jump_interpolation(
        x_in: xr.Dataset, 
        x_dim: str="time", 
        factor: float=0.001, 
        interpolation: Literal["fill-forward", "linear"] = "fill-forward",
    ) -> xr.Dataset:
    """Make the interpolation safe by adding a coordinate just before each 
    x-value (except the first vaue). The distance between the new and the next
    point are calculated as a fraction of the previous distance between
    neighboring points. The corresponding y-values are first set to NaN and then
    interpolated based on the interpolation method.

    Parameters
    ----------
    x_in : xr.Dataset
        The input dataset which contains a coordinate (x) and a data variable
        (y)
    x_dim : str, optional
        The name of the x coordinate, by default "time"
    factor : float, optional
        The distance between the newly added points and the following existing
        points on the x-scale, by default 1e-4
    interpolation : Literal["fill-forward", "linear"], optional
        The interpolation method. In addition to 'fill-forward' and 'linear',
        any method give in `xr.interpolate_na` can be chosen, by default
        "fill-forward"

    Returns
    -------
    xr.Dataset
        The interpolated dataset
    """
    xs = x_in.coords[x_dim]

    # calculate x values that are located just a little bit smaller than the xs
    # where "just a little bit" is defined by the distance to the previous x
    # and a factor. This way the scale of the observations should not matter
    # and even very differently sized x-steps should be interpolated correctly
    fraction_before_xs = (
        xs.isel({x_dim:range(1, len(xs))})
        - xs.diff(dim=x_dim) * factor
    )

    # create a sorted time vector
    xs = sorted([*fraction_before_xs.values, *xs.values])

    # add new time indices with NaN values 
    x_in_reindexed = x_in.reindex({x_dim:xs})

    if interpolation == "fill-forward":
        # then fill nan values with the previous value (forward-fill)
        x_in_interpolated = x_in_reindexed.ffill(dim=x_dim, limit=1)

    else:
        x_in_interpolated = x_in_reindexed.interpolate_na(dim=x_dim, method="linear")

    return x_in_interpolated


def smoothed_interpolation(
    x_in: xr.Dataset, 
    x_dim: str="time", 
    factor: float=0.001, 
    sigma: int = 20,
) -> xr.Dataset:
    """Smooth the interpolation by first creating a dense x vector and forward
    filling all ys. Following this the values are smoothed by a gaussian filter.

    Parameters
    ----------
    x_in : xr.Dataset
        The input dataset which contains a coordinate (x) and a data variable
        (y)
    x_dim : str, optional
        The name of the x coordinate, by default "time"
    factor : float, optional
        The distance between the newly added points and the following existing
        points on the x-scale, by default 1e-4

    Returns
    -------
    xr.Dataset
        The interpolated dataset
    """
    xs = x_in.coords[x_dim]
    assert factor > 0, "Factor must be larger than zero, to ensure correct ordering"
    
    xs_extra = np.arange(xs.values.min(), xs.values.max()+factor, step=factor)
    xs_ = np.sort(np.unique(np.concatenate([xs.values, xs_extra])))

    # add new time indices with NaN values 
    x_in_reindexed = x_in.reindex({x_dim:xs_})

    # then fill nan values with the previous value (forward-fill)
    x_in_interpolated = x_in_reindexed.ffill(dim=x_dim)

    # Apply Gaussian smoothing
    sigma = 20  # Adjust sigma for desired smoothness
    for k in x_in_interpolated.data_vars.keys():
        y = x_in_interpolated[k]
        y_smoothed = gaussian_filter1d(y.values, sigma, axis=list(y.dims).index(x_dim))

        x_in_interpolated[k].values = y_smoothed

    return x_in_interpolated


def radius_interpolation(
    x_in: xr.Dataset, 
    x_dim: str="time", 
    radius: float=0.1, 
    num_points: int=10,
    rectify=True
) -> xr.Dataset:
    """Smooth the interpolation by first creating a dense x vector and forward
    filling all ys. Following this the values are smoothed by a gaussian filter.

    WARNING! It is very pretty but does not work with diffrax

    Parameters
    ----------
    x_in : xr.Dataset
        The input dataset which contains a coordinate (x) and a data variable
        (y)
    x_dim : str, optional
        The name of the x coordinate, by default "time"
    radius : float, optional
        The radius of the quarter-circle to curve the jump transition. 
        By default 0.1
    num_points : int, optional
        The number of points to interpolate each jump with. Default: 10
    rectify : bool
        Whether the input should be converted to a stepwise pattern. Default 
        is True. This is typically applied if an unprocessed signal is included.
        E.g. the signal was observed y_i 
        
    Returns
    -------
    xr.Dataset
        The interpolated dataset
    """
    x = x_in.coords[x_dim] 
    assert radius <= np.diff(np.unique(x)).min() / 2

    if rectify:
        x = np.concatenate([[x[0]], *[[x_i-0.00, x_i] for x_i in x[1:]]])

    data_arrays = []
    for k in x_in.data_vars.keys():
        y = x_in[k]
        if rectify:
            yvals = np.concatenate([*[[y_i, y_i] for y_i in y[:-1]], [y[-1]]])
        else:
            yvals = y.values

        x_interpolated = [np.array(x[0],ndmin=1)]
        y_interpolated = [np.array(yvals[0], ndmin=2)]
        for i in range(0, len(x) - 1):
            x_, y_, = curve_jumps(x, yvals, i, r=radius, n=num_points)

            x_interpolated.append(x_)
            y_interpolated.append(y_)
    
        x_interpolated = np.concatenate(x_interpolated)
        x_uniques = np.where(np.diff(x_interpolated) != 0)
        x_interpolated = x_interpolated[x_uniques]

        y_interpolated = np.row_stack(y_interpolated)[x_uniques]
        
        coords = {x_dim: x_interpolated}
        coords.update({d: y.coords[d].values for d in y.dims if d != x_dim})

        y_reindexed = xr.DataArray(
            y_interpolated, 
            coords=coords,
            name=y.name
        )
        data_arrays.append(y_reindexed)

    x_in_interpolated = xr.combine_by_coords(data_objects=data_arrays)
    
    # there will be nans if the data variables have different steps
    # x_in_interpolated = x_in_interpolated.interpolate_na(
    #     dim="time", method="linear"
    # )

    return x_in_interpolated # type: ignore


def curve_jumps(x, y, i, r, n):
    x_i = x[i] # jump start
    y_i = y[i] # jump start
    y_im1 = y[i-1] # jump start
    y_ip1 = y[i+1] # jump end

    def circle_func(x, r, a): 
        # using different radii does not work, because this would also require different x_values
        arc = r**2 - (x - a)**2
        return np.sqrt(np.where(arc >= 0, arc, 0))
    
    # end of jump
    dy_im1 = y_i - y_im1 # jump difference to previous point
    if np.all(dy_im1 == 0):
        dyj = y_ip1 - y_i
        sj = np.where(np.abs(dyj) > r, np.sign(dyj), dyj / 2 / r) # direction and size of the jump, scaled by the radius
        # sj = np.clip((y_ip1 - y_i)/r/2, -1, 1) # direction and size of the jump, scaled by the radius
        xc = np.linspace(x_i-r, x_i-r*0.001, num=n)
        yc = y_i + (np.einsum("k,x->xk",  -sj, circle_func(x=xc, r=r, a=x_i-r)) + sj * r)
    else:
        dyj = y_i - y_im1
        sj = np.where(np.abs(dyj) > r, np.sign(dyj), dyj / 2 / r) # direction and size of the jump, scaled by the radius
        xc = np.linspace(x_i+ r*0.001, x_i+r, num=n)
        yc = y_i + (np.einsum("k,x->xk",  sj, circle_func(x=xc, r=r, a=x_i+r)) - sj * r)
    
    return xc, yc


def rect_interpolation(
    x_in: xr.Dataset, 
    x_dim: str="time", 
):  
    """Use diffrax' rectilinear_interpolation. To add values and interpolate
    one more step after the end of the timeseries
    """
    data_arrays = []
    data_vars = tuple(x_in.data_vars.keys())
    dataset_dims = tuple(x_in.dims.keys())

    for k in data_vars:
        v_orig = x_in[k]
        v = v_orig.transpose(x_dim, ...)
        x = v[x_dim].values
        y = v.values
        x_dim_loc = v.dims.index(x_dim)

        ts_ = np.concatenate([x, np.array(x[-1]+1, ndmin=1)]) 
        if y.ndim == 1:
            ys_ = np.concatenate([y, np.array(y[-1], ndmin=1)])
        elif y.ndim == 2:
            if x_dim_loc == 0:
                ys_ = np.row_stack([y, np.expand_dims(y[-1], axis=0)]) 
        elif y.ndim == 3:
            if x_dim_loc == 0:
                ys_ = np.row_stack([y, np.expand_dims(y[-1], axis=0)])
            else:
                raise ValueError(f"{x_dim_loc} should be the first dimension.")
        else:
            raise NotImplementedError(
                "Dimensions of interpolation > 2 or 0 are not implemented"
            )

        xs, ys = rectilinear_interpolation(ts=ts_, ys=ys_) # type:ignore
        # if y.ndim <= 2:
        #     xs = xs
        # else:
        #     xs = xs[0] # type:ignore

        coords = {x_dim: xs}
        coords.update({d: v.coords[d].values for d in v.dims if d != x_dim})
        # coords = {d:coords[d] for d in v.dims}

        y_reindexed = xr.DataArray(
            ys, 
            coords=coords,
            name=v.name
        )
        y_reindexed = y_reindexed.transpose(*v_orig.dims)

        data_arrays.append(y_reindexed)

    x_in_interpolated = xr.combine_by_coords(data_objects=data_arrays)
    x_in_interpolated = x_in_interpolated.transpose(*dataset_dims)


    return x_in_interpolated[[*dataset_dims, *data_vars]]


import warnings

import numpy as np
import xarray as xr
import arviz as az
from arviz.sel_utils import xarray_var_iter
from matplotlib import pyplot as plt
import matplotlib as mpl

from pymob.utils.plot_helpers import plot_hist, plot_loghist

def cluster_chains(posterior, deviation="std"):
    assert isinstance(posterior, (xr.DataArray, xr.Dataset))
    chain_means = posterior.mean(dim="draw")
    if deviation == "std":
        chain_dev = posterior.std(dim="draw")
    elif "frac:" in deviation:
        _, frac = deviation.split(":")
        chain_dev = chain_means * float(frac)
    else:
        raise ValueError("Deviation method not implemented.")    

    global cluster_id
    cluster_id = 1
    cluster = [cluster_id] * len(posterior.chain)
    unclustered_chains = posterior.chain.values

    def recurse_clusters(unclustered_chains):
        global cluster_id
        compare = unclustered_chains[0]
        new_cluster = []
        for i in unclustered_chains[1:]:
            a = chain_means.sel(chain=compare) + chain_dev.sel(chain=compare) > chain_means.sel(chain=i)
            b = chain_means.sel(chain=compare) - chain_dev.sel(chain=compare) < chain_means.sel(chain=i)
            isin_dev = (a * b).all()

            if isinstance(isin_dev, xr.Dataset):
                isin_dev = isin_dev.to_array().all()

            if not isin_dev:
                cluster[i] = cluster_id + 1
                new_cluster.append(i)

        cluster_id += 1
        if len(new_cluster) == 0:
            return

        recurse_clusters(new_cluster)
    
    recurse_clusters(unclustered_chains)

    return cluster


def rename_extra_dims(df, extra_dim_suffix="_dim_0", new_dim="new_dim", new_coords=None):
    # TODO: COuld be used for numypro backend for fixing posterior indexes
    df_ = df.copy()
    data_vars = list(df_.data_vars.keys())

    # swap dimension names for all dims that have the suffix 
    new_dims = {}
    for dv in data_vars:
        old_dim = f"{dv}{extra_dim_suffix}"
        if df_.dims[old_dim] == 1:
            df_[dv] = df_[dv].squeeze(old_dim)
        else:
            new_dims.update({old_dim: new_dim})

    df_ = df_.swap_dims(new_dims)

    # assign coords to new dimension
    df_ = df_.assign_coords({new_dim: new_coords})
    
    # drop renamed coords
    df_ = df_.drop([f"{dv}{extra_dim_suffix}" for dv in data_vars])

    return df_



# plot loghist
def plot_posterior_samples(posterior, col_dim=None, log=True, hist_kwargs={}):
    if log:
        hist = plot_loghist
    else:
        hist = plot_hist

    parameters = list(posterior.data_vars.keys())
    samples = posterior.stack(sample=("chain", "draw"))

    fig = plt.figure(figsize=(5, len(parameters)*2))
    fig.subplots_adjust(right=.95, top=.95, hspace=.25)

    gs = fig.add_gridspec(len(parameters), 1)

    for i, key in enumerate(parameters):
        postpar = samples[key]
        if col_dim in postpar.dims:
            col_coords = postpar[col_dim]
            gs_par = gs[i, 0].subgridspec(1, len(col_coords))
            axes = gs_par.subplots()

            for ax, coord in zip(axes, col_coords):
                hist(
                    x=postpar.sel({col_dim: coord}), 
                    name=f"${key}$ {str(coord.values)}",
                    ax=ax,
                    **hist_kwargs
                )

        else:
            gs_par = gs[i, 0].subgridspec(1, 1)
            ax = gs_par.subplots()

            hist(
                x=postpar, 
                name=f"${key}$",
                ax=ax,
                **hist_kwargs
            )


    return fig


def bic(idata: az.InferenceData):
    """calculate the BIC for az.InferenceData. The function will average over
    all samples from the markov chain
    """
    log_likelihood = idata.log_likelihood.mean(("chain", "draw")).sum().to_array().sum()
    k = idata.posterior.mean(("chain", "draw")).count().to_array().sum()

    vars = [i.split("_")[0] for i in list(idata.log_likelihood.data_vars.keys())]
    n = 0
    for v in vars:
        if v in idata.observed_data:
            n += (~idata.observed_data[v].isnull()).sum()
        elif v + "_obs" in idata.observed_data:
            n += (~idata.observed_data[v + "_obs"].isnull()).sum()
        else:
            raise IndexError(f"Variable {v} or {v+'_obs'} not in idata")

    # n = (~idata.observed_data[vars].isnull()).sum().to_array().sum()

    bic = float(k * np.log(n) - 2 * log_likelihood)
    msg = str(
        "Bayesian Information Criterion (BIC)"
        "\n===================================="
        f"\nParameters: {int(k)}"
        f"\nData points: {int(n)}"
        f"\nLog-likelihood: {float(log_likelihood)}"
        f"\nBIC: {bic}"
    )

    return msg, bic


def add_cluster_coordinates(idata, deviation="std"):
    cluster = cluster_chains(idata.posterior, deviation=deviation)
    idata = idata.assign_coords(cluster=("chain", cluster))
    return idata


def format_parameter(par, subscript_sep="_", superscript_sep="__", textwrap="\\text{}"):
    super_pos = par.find(superscript_sep)
    sub_pos = par.find(subscript_sep)
    
    scripts = sorted(zip([sub_pos, super_pos], [subscript_sep, superscript_sep]))
    scripts = [sep for pos, sep in scripts if pos > -1]


    def wrap_text(substr):
        if len(substr) == 1:
            substring_fmt = f"{substr}"
        else:
            substring_fmt = textwrap.replace("{}", "{{{}}}").format(substr)
    
        return f"{{{substring_fmt}}}"

    formatted_string = "$"
    for i, sep in enumerate(scripts):
        substr, par = par.split(sep, 1)
        substring_fmt = wrap_text(substr=substr)

        math_sep = "_" if sep == subscript_sep else "^"

        formatted_string += substring_fmt + math_sep

    formatted_string += wrap_text(par) + "$"

    return formatted_string


def create_table(posterior, error_metric="hdi", vars={}, nesting_dimension=None):
    """The function is not ready to deal with any nesting dimensionality
    and currently expects the 2-D case
    """
    parameters = list(posterior.data_vars.keys())

    tab = az.summary(posterior, fmt="xarray", kind="stats", stat_focus="mean", hdi_prob=0.94)
    if len(vars) > 0:
        tab = tab[vars.keys()]

    tab = tab.rename(vars)

    if error_metric == "sd":
        arrays = []
        for par in parameters:
            par_formatted = tab.sel(metric=["mean", "sd"])[par]\
                .round(3)\
                .astype(str).str\
                .join("metric", sep=" ±")
            arrays.append(par_formatted)

        formatted_tab = xr.combine_by_coords(arrays).to_dataframe().T

        formatted_parameters = []
        for idx in formatted_tab.index:
            formatted_parameters.append(format_parameter(idx))


        formatted_tab.index = formatted_parameters
        return formatted_tab    
    

    elif error_metric == "hdi":
        stacked_tab = tab.sel(metric=["mean", "hdi_3%", "hdi_97%"])\
            .assign_coords(metric=["mean", "hdi 3%", "hdi 97%"])\
            .stack(col=(nesting_dimension, "metric"))\
            .round(2)
        formatted_tab = stacked_tab.to_dataframe().T.drop(index=[nesting_dimension, "metric"])

        return formatted_tab
    
    else:
        raise NotImplementedError("Must use one of 'sd' or 'hdi'")
    

def filter_not_converged_chains(idata, deviation=1.05):
    posterior = idata.posterior
    log_likelihood = idata.log_likelihood
    log_likelihood_summed = log_likelihood.to_array("obs")
    log_likelihood_summed = log_likelihood_summed.sum(("time", "id", "obs"))

    # filter non-converged parameter estimates
    likelihood_mask = (
        # compares the mean of the summed log likelihood of a given chain
        log_likelihood_summed.mean("draw") > 
        # to the maximum of all chain means times a factor
        log_likelihood_summed.mean("draw").max() * deviation
    )
    posterior_filtered = posterior.where(likelihood_mask, drop=True)
    log_likelihood_filtered = log_likelihood.where(likelihood_mask, drop=True)

    idata = az.InferenceData(
        posterior=posterior_filtered, 
        log_likelihood=log_likelihood_filtered,
        observed_data=idata.observed_data,
    )

    return idata


def log(msg, out, newlines=1, mode="a"):
    with open(out, mode) as f:
        print(msg, file=f, end="\n")
        for _ in range(newlines):
            print("", file=f, end="\n")


def evaluate_posterior(sim, nesting_dimension, n_samples=10_000, vars_table={}, 
                       seed=1, save=True, show=False):
    """The function is not ready to deal with any nesting dimensionality 
    and currently expects the 2-D case
    """
    rng = np.random.default_rng(seed)
    idata = sim.inferer.idata
    # idata.posterior = idata.posterior.chunk(chunks={"draw":100}).load()
    # idata.log_likelihood = idata.log_likelihood.chunk(chunks={"draw":100})
    n_subsample = min(
        int(n_samples / idata.posterior.sizes["chain"]), 
        idata.posterior.sizes["draw"]
    )

    if n_subsample < 250:
        warnings.warn(
            "The number of samples drawn from each chain for the pairplot "
            f"({n_subsample}) may be too small to be representative. "
            "Consider increasing n_samples."
        )

    subsamples = rng.choice(idata.posterior.draw, n_subsample, replace=False)
    idata.posterior = idata.posterior.sel(draw=subsamples)
    idata.log_likelihood = idata.log_likelihood.sel(draw=subsamples)

    log_likelihood_summed = idata.log_likelihood.to_array("obs")
    log_likelihood_summed = log_likelihood_summed.sum(("time", "id", "obs"))

    print(az.summary(idata.posterior))

    table = create_table(
        posterior=idata.posterior, 
        error_metric="hdi",
        vars=vars_table,
        nesting_dimension=nesting_dimension,
    )
    table = table.rename(columns={"hdi 3%": "hdi 3\\%", "hdi 97%": "hdi 97\\%"})
    table = table.rename(columns={c: c.capitalize() for c in table.columns.levels[0]})
    table.columns = table.columns.set_names(["Parameters", ""])
    table.index = [format_parameter(i) for i in list(table.index)]
    table_latex = table.to_latex(
        float_format="%.2f",
        caption=(
            "Parameter estimates and posterior highest densitiy intervals (HDI) "+
            f"of the {sim.case_study}__{sim.scenario} model. The HDI "+
            "contains 94\% of the probable parameter values given the data."
        ),
        label=f"tab:parameters-{sim.case_study}__{sim.scenario}"
    )

    # bic 
    msg, _ = bic(idata)
    if save:
        log(table_latex, out=f"{sim.output_path}/parameter_table.tex", mode="w")
        log(msg=msg, out=f"{sim.output_path}/bic.md", mode="w")

    if show:
        print(table)
        print(msg)

    fig_param = plot_posterior_samples(
        idata.posterior, 
        col_dim=nesting_dimension, 
        log=True,
        hist_kwargs = dict(hdi=True, bins=20)
    )
    fig_param.set_size_inches(12, 30)

    if save:
        fig_param.savefig(f"{sim.output_path}/multichain_parameter_estimates.jpg")

    if show:
        plt.show()
    else:
        plt.close()

    def plot_pairs(posterior, likelihood):
        parameters = list(posterior.data_vars.keys())

        N = len(parameters)
        parameters_ = parameters.copy()
        fig = plt.figure(figsize=(3*N, 3*(N+1)))
        gs = fig.add_gridspec(N, N+1, width_ratios=[1]*N+[0.2])
        

        i = 0
        while len(parameters_) > 0:
            par_x = parameters_.pop(0)
            hist_ax = gs[i,i].subgridspec(1, 1).subplots()
            plot_hist(
                posterior[par_x].stack(sample=("chain", "draw")), 
                ax=hist_ax, decorate=False, bins=20
            )
            hist_ax.set_title(par_x)
            for j, par_y in enumerate(parameters_, start=i+1):
                ax = gs[j,i].subgridspec(1, 1).subplots()

                scatter = ax.scatter(
                    posterior[par_x], 
                    posterior[par_y], 
                    c=likelihood, 
                    alpha=0.25,
                    s=10,
                    cmap=mpl.colormaps["plasma_r"]
                )

                if j != len(parameters)-1:
                    ax.set_xticks([])
            
                ax.set_xlabel(par_x)            
                ax.set_ylabel(par_y)

            i += 1

        # ax_colorbar = gs[:,N].subgridspec(1, 1).subplots()
        # fig.colorbar(scatter, cax=ax_colorbar)
        return fig

    for coord in idata.posterior[nesting_dimension].values:
        print("=" * len(coord))
        print(coord.capitalize())
        print("=" * len(coord))
        az.plot_trace(idata.posterior.sel({nesting_dimension:coord}))
        
        if save:
            plt.savefig(f"{sim.output_path}/multichain_pseudo_trace_{coord}.jpg")

        if show:
            plt.show()
        else:
            plt.close()

        fig = plot_pairs(
            posterior=idata.posterior.sel({nesting_dimension:coord}), 
            likelihood=log_likelihood_summed,
        )

        if save:
            fig.savefig(f"{sim.output_path}/multichain_pairs_{coord}.jpg")

        if show:
            plt.show()
        else:
            plt.close()



def plot_pair(posterior, likelihood, parameters, axes=None):

    likelihood_vars = list(likelihood.data_vars.keys())
    N = len(likelihood_vars)

    if axes is None:
        fig, axes = plt.subplots(1, N, figsize=(3+(2*(N-1)), 3), sharey=True, sharex=True)
    
    if isinstance(likelihood, xr.DataArray):
        if likelihood.name is None:
            likelihood.name = "joint_likelihood"
        likelihood = likelihood.to_dataset()
    

    par_a, par_b = parameters

    post_a = posterior[par_a]
    post_b = posterior[par_b]

    # always put the smaller parameter on the x-axis
    if post_a.ndim <= post_b.ndim:
        post_x = post_a
        post_y = post_b
    else:
        post_x = post_b.stack(sample=("chain", "draw"))
        post_y = post_a.stack(sample=("chain", "draw"))

    loglik = likelihood.stack(sample=("chain","draw"))
    extra_dims_y = [d for d in post_y.dims if d not in post_x.dims]
    norm = mpl.colors.Normalize(vmin=(likelihood).to_array().min(), vmax=(likelihood).to_array().max())
    cmap = mpl.colormaps["plasma"]

    for likvar, ax in zip(likelihood_vars, axes):
        ll = loglik[likvar]
        ax.set_title(likvar)
        ax.set_xlabel(post_x.name)
        ax.set_ylabel(post_y.name)
        for dim in extra_dims_y:
            for coord in post_y.coords[dim].values:
                if dim in ll.dims:
                    ll_coord = ll.sel({dim:coord})
                else:
                    ll_coord = ll

                scatter = ax.scatter(
                    post_x, 
                    post_y.sel({dim:coord}), 
                    c=ll_coord, 
                    alpha=0.25,
                    marker=f"${coord}$",
                    s=40,
                    cmap=cmap,
                    norm=norm
                )

                # ax.annotate()
        
    cbar = fig.colorbar(
        mappable=mpl.cm.ScalarMappable(norm=norm, cmap=cmap), 
        # if the same axis is used it is plot next to
        # the plot depending on the location option
        # ax=ax,
        # location="right",
        orientation="vertical",

        # if another axis is used, the fraction should be increased to 1.0
        ax=ax,
        fraction=0.1,
        pad=0.05,
        # aspect=20, # parameter is not so important. It thins the colorbar
        label="negative log-likelihood"
    )
    fig.tight_layout()

    # ax_colorbar = gs[:,N].subgridspec(1, 1).subplots()
    return fig


import numpy as np

def assert_no_infs_in_dataset(dataset):
    infs = bool((dataset == np.inf).sum().to_array().any())
    if infs:
        raise ValueError("Dataset contained infinity values")

def assert_no_nans_in_dataset(dataset):
    nans = bool(dataset.isnull().sum().to_array().any())
    if nans:
        raise ValueError("Dataset contained NaN values")




from typing import Literal
import xarray as xr


# legacy imports. These functions are now included in pymob.solvers.base
from pymob.solvers.base import mappar, smoothed_interpolation, jump_interpolation, rect_interpolation
from pymob.solvers.analytic import solve_analytic_1d
from pymob.solvers.scipy import solve_ivp_1d





"""
TODO: Case study tests should be outsourced to the different tests. Nevertheless
the tests should be run when pymob is tested, but not in here but in the 
case studies.
"""


import numpy as np
import pytest

from tests.fixtures import (
    init_simulation_casestudy_api,
    init_case_study_and_scenario,
    init_guts_casestudy_constant_exposure,
    init_guts_casestudy_variable_exposure,
    init_bufferguts_leo_casestudy,
    init_bufferguts_casestudy,
)


def test_multiple_case_study_import():
    sim = init_simulation_casestudy_api()
    sim = init_guts_casestudy_constant_exposure()
    sim = init_bufferguts_casestudy()
    sim = init_bufferguts_casestudy()
    sim = init_simulation_casestudy_api()
    sim = init_case_study_and_scenario(
        case_study="tktd_rna_pulse", 
        scenario="rna_pulse_3_6c_substance_specific"
    )





def test_bufferguts_variable_exposure():
    pytest.skip()
    sim = init_bufferguts_casestudy()

    # sim.use_symbolic_solver(do_compile=True)
    # evaluator = sim.dispatch(theta={})
    # evaluator()
    # sol_symbolic = evaluator.results

    sim.use_discrete_solver()
    evaluator = sim.dispatch(theta={})
    evaluator()
    sol_discrete = evaluator.results

    sim.use_jax_solver()
    sim.dispatch_constructor(rtol=1e-7, atol=1e-8, max_steps=1e6)
    evaluator = sim.dispatch(theta={})
    evaluator()
    sol_numerical = evaluator.results

    # make sure errors are small between exact solution and numerical solution
    # the errors come from:
    # a) integrating not exactly to t_eq
    # b) using a numerical switch in the ODE solution.
    diff = (
        sol_numerical.sel(time=np.arange(0,sim.t_max))
        - sol_discrete.sel(time=np.arange(0,sim.t_max))
    )[["B", "D", "H", "S"]]
    max_delta = np.abs(diff).max().to_array()
    np.testing.assert_array_less(max_delta, [1e-2, 1e-8, 1e-8])


    axes = sim._plot.plot_multiexposure(sol_numerical, vars=["exposure", "B", "D", "H", "S"], color="tab:blue", label_prefix="ODE")
    axes = sim._plot.plot_multiexposure(sol_discrete, vars=["exposure", "B", "D", "H", "S"], axes=axes, color="tab:red", linestyle="--", label_prefix="exact")
    fig = axes[0].figure
    fig.savefig(f"{sim.output_path}/solution_comparison.png")


def test_bufferguts_conditional_binomial():
    pytest.skip()
    sim = init_case_study_and_scenario(case_study="bufferguts", scenario="simulate_lab_experiment")
    
    sim.config.inference_numpyro.user_defined_error_model = "conditional_survival_error_model"
    sim.dispatch_constructor()
    sim.set_inferer("numpyro")

    sim.config.inference_numpyro.kernel = "nuts"
    sim.config.inference_numpyro.svi_iterations = 1000
    sim.config.inference_numpyro.chains = 1
    sim.inferer.run(print_debug=False, render_model=False)

    sim.inferer.idata.posterior_predictive

    sim.inferer.idata.posterior
    # TODO: Test output somehow



def test_bufferguts_multinomial():
    pytest.skip()
    sim = init_case_study_and_scenario(case_study="bufferguts", scenario="simulate_lab_experiment_multinomial")
    
    sim.config.inference_numpyro.user_defined_error_model = "conditional_survival_error_model"
    sim.dispatch_constructor()
    sim.set_inferer("numpyro")
    
    sim.config.inference_numpyro.kernel = "map"
    sim.config.inference_numpyro.svi_iterations = 10
    sim.config.inference_numpyro.chains = 1
    sim.inferer.run()

    sim.inferer.idata


def test_bufferguts_hybrid_solution():
    sim = init_bufferguts_casestudy()

    sim.use_hybrid_solver()
    evaluator = sim.dispatch(theta={})
    evaluator()
    sol_symbolic = evaluator.results

    sim.use_jax_solver()
    sim.dispatch_constructor(rtol=1e-7, atol=1e-8, max_steps=1e6)
    evaluator = sim.dispatch(theta={})
    evaluator()
    sol_numerical = evaluator.results

    # make sure errors are small between exact solution and numerical solution
    # the errors come from:
    # a) integrating not exactly to t_eq
    # b) using a numerical switch in the ODE solution.
    diff = (
        sol_numerical.sel(time=np.arange(0,sim.t_max))
        - sol_symbolic.sel(time=np.arange(0,sim.t_max))
    )[["B", "D", "H", "S"]]
    max_delta = np.abs(diff).max().to_array()
    # large errors are partially due to the solver running at 64 bit precision 
    # (exact to 8 digits)
    np.testing.assert_array_less(max_delta, [0.25, 0.25, 0.25, 0.25])


    axes = sim._plot.plot_multiexposure(sol_numerical, vars=["exposure", "B", "D", "H", "S"], color="tab:blue", label_prefix="ODE")
    axes = sim._plot.plot_multiexposure(sol_symbolic, vars=["exposure", "B", "D", "H", "S"], axes=axes, color="tab:red", linestyle="--", label_prefix="exact")
    fig = axes[0].figure
    fig.savefig(f"{sim.output_path}/solution_comparison.png")

def test_bufferguts_leo():
    sim = init_bufferguts_leo_casestudy()
    sim.initialize(["Beta-Cyfluthrin_M-051896_acute_oral.xlsx", "Exposure", "BufferGUTS_SD"])
    sim.setup_numpyro_inferer()

    e = sim.dispatch({})
    e()
    e.results

    sim.inferer.run()
    sim.inferer.store_results()
    fig = sim.plot_posterior_predictions()
    fig.savefig(f"{sim.output_path}/posterior_model_fits.png")


if __name__ == "__main__":
    import sys
    import os
    sys.path.append(os.getcwd())
    pass
    # test_multiple_case_study_import()
    # test_constant_exposure()
    # test_variable_exposure()
    # test_bufferguts_variable_exposure()
    # test_bufferguts_hybrid_solution()



import numpyro
import jax.numpy as jnp
from numpyro.distributions import Normal, TransformedDistribution
from numpyro.distributions import transforms

scipy_to_numpyro = {
    "deterministic": (numpyro.deterministic, {"value": "value"}),
    # Continuous Distributions
    "norm": (numpyro.distributions.Normal, {"loc": "loc", "scale": "scale", "low": "low", "high": "high"}),
    "normal": (numpyro.distributions.Normal, {"loc": "loc", "scale": "scale", "low": "low", "high": "high"}),
    "halfnorm": (numpyro.distributions.HalfNormal, {"scale": "scale", "high":"high"}),
    "halfnormal": (numpyro.distributions.HalfNormal, {"scale": "scale", "high":"high"}),
    "expon": (numpyro.distributions.Exponential, {"scale": "rate", "high": "high"}),  # rate = 1/scale
    "exponential": (numpyro.distributions.Exponential, {"scale": "rate", "high": "high"}),  # rate = 1/scale
    "uniform": (numpyro.distributions.Uniform, {"loc": "low", "scale": "high"}),  # high = loc + scale
    "beta": (numpyro.distributions.Beta, {"a": "concentration1", "b": "concentration0"}),
    "gamma": (numpyro.distributions.Gamma, {"a": "concentration", "scale": "rate", "low": "low", "high": "high"}),  # rate = 1/scale
    "lognorm": (numpyro.distributions.LogNormal, {"scale": "loc", "s": "scale", "loc": "loc", "low": "low", "high": "high"}),
    "lognormal": (numpyro.distributions.LogNormal, {"scale": "loc", "s": "scale", "loc": "loc", "low": "low", "high": "high"}),
    "chi2": (numpyro.distributions.Chi2, {"df": "df", "low": "low", "high": "high"}),
    "pareto": (numpyro.distributions.Pareto, {"b": "scale", "scale": "alpha", "low": "low", "high": "high"}),
    "t": (numpyro.distributions.StudentT, {"df": "df", "loc": "loc", "scale": "scale", "low": "low", "high": "high"}),
    "cauchy": (numpyro.distributions.Cauchy, {"loc": "loc", "scale": "scale", "low": "low", "high": "high"}),
    "gumbel_r": (numpyro.distributions.Gumbel, {"loc": "loc", "scale": "scale", "low": "low", "high": "high"}),
    "gumbel_l": (lambda loc, scale: numpyro.distributions.Gumbel(loc=-loc, scale=scale), {}), 
    "laplace": (numpyro.distributions.Laplace, {"loc": "loc", "scale": "scale", "low": "low", "high": "high"}),
    "logistic": (numpyro.distributions.Logistic, {"loc": "loc", "scale": "scale", "low": "low", "high": "high"}),
    "multivariate_normal": (numpyro.distributions.MultivariateNormal, {"mean": "loc", "cov": "covariance_matrix"}),
    "dirichlet": (numpyro.distributions.Dirichlet, {"alpha": "concentration"}),
    
    # Discrete Distributions
    "binom": (numpyro.distributions.Binomial, {"n": "total_count", "p": "probs"}),
    "binomial": (numpyro.distributions.Binomial, {"n": "total_count", "p": "probs"}),
    "bernoulli": (numpyro.distributions.Bernoulli, {"p": "probs"}),
    "geom": (numpyro.distributions.Geometric, {"p": "probs"}),
    "poisson": (numpyro.distributions.Poisson, {"mu": "rate"}),
    "nbinom": (numpyro.distributions.NegativeBinomialProbs, {"n": "total_count", "p": "probs"}),
    "multinomial": (numpyro.distributions.Multinomial, {"n": "total_count", "p": "probs"}),
    "categorical": (numpyro.distributions.Categorical, {"p": "probs"}),
    
}

import jax.numpy as jnp
from numpyro.distributions import Normal, TransformedDistribution, TruncatedNormal
from numpyro.distributions import transforms


def transform(transforms, x):
    for part in transforms:
        x = part(x)
    return x

def inv_transform(transforms, y):
    for part in transforms[::-1]:
        y = part.inv(y)
    return y



# LogNormal Transformation
def LogNormalTrans(loc, scale, low=None, high=None):
    _transforms = [
        transforms.AffineTransform(loc=jnp.log(loc), scale=scale),
        transforms.ExpTransform(),
    ]

    if high is not None:
        high = inv_transform(_transforms, high)

    if low is not None:
        low = inv_transform(_transforms, low)

    base_distribution=TruncatedNormal(loc=0, scale=1, low=low, high=high)
    return TransformedDistribution(
        base_distribution=base_distribution,
        transforms=_transforms
    )

# LogNormal Transformation
def NormalTrans(loc, scale):
    return TransformedDistribution(
        base_distribution=Normal(0, 1),
        transforms=[
            transforms.AffineTransform(loc=loc, scale=scale),
        ]
    )

# Exponential Transformation
def ExponentialTrans(rate):
    return TransformedDistribution(
        base_distribution=Normal(0, 1),
        transforms=[
            transforms.AffineTransform(loc=0, scale=1/rate),
            transforms.ExpTransform(),
        ]
    )

# Uniform Transformation
def UniformTrans(low, high):
    return TransformedDistribution(
        base_distribution=Normal(0, 1),
        transforms=[
            transforms.SigmoidTransform(),
            transforms.AffineTransform(loc=low, scale=high - low)
        ]
    )


# Half-Normal Transformation
def HalfNormalTrans(scale):
    return TransformedDistribution(
        base_distribution=Normal(0, 1),
        transforms=[
            transforms.AffineTransform(loc=0, scale=scale),
            transforms.AbsTransform()
        ]
    )

transformed_dist_map = {
    numpyro.distributions.LogNormal: LogNormalTrans,
    numpyro.distributions.Exponential: ExponentialTrans,
    numpyro.distributions.Uniform: UniformTrans,
    numpyro.distributions.HalfNormal: HalfNormalTrans,
    numpyro.distributions.Normal: NormalTrans,
}


from pymob.solvers.base import mappar

def solve_analytic_1d(model, parameters, dimensions, coordinates, data_variables, seed=None):
    """Solves an anlytic function for all coordinates in the first data 
    dimension of the model

    - parameters: define the model
    - y0: set the initial values of the ODE states
    - coordinates: are needed to know over which values to integrate
    - seed: In case stochastic processes take place inside the model this is necessary
    
    In order to make things explicit, all information which is needed by the
    model needs to be specified in the function signature. 
    This also makes the solvers functionally oriented, a feature that helps the
    usability of models accross inference frameworks. Where functions should not
    have side effects.

    Additionally, passing arguments via the signature makes it easier to write
    up models in a casual way and only later embed them into more regulated
    structures such as pymob

    """
    
    model_args = mappar(
        model, 
        parameters["parameters"], 
        exclude=["t", "x"] + dimensions,
        to="dict"
    )
    
    x = coordinates[dimensions[0]]

    model_results = []
    results = model(x, **model_args)
    model_results.append(results)
    
    return {data_var:y for data_var, y in zip(data_variables, model_results)}



import numpy as np
import ast
from typing import (
    Dict,
    Tuple,
    Mapping,
    Callable,
    Iterable,
    Optional,
    Any,
    List,
    Protocol,
    Literal
)
from abc import ABC, abstractmethod
import warnings

from matplotlib import pyplot as plt
import matplotlib as mpl
import arviz as az
import itertools as it

from pymob.simulation import SimulationBase
from pymob.sim.parameters import Param, RandomVariable, Expression
from pymob.sim.config import Datastructure
from pymob.utils.config import lookup_from

class Errorfunction(Protocol):
    def __call__(
        self, 
        theta: Dict[str, List[float]], 
    ) -> Any:
        ...

class Distribution:
    """The distribution is a pre-initialized distibution with human friendly 
    interface to construct a distribution in an arbitrary backend.

    The necessary adjustments to make the distribution backend specific
    are done by passing more context to Distribution class to the
    _context variable
    """
    distribution_map: Dict[str,Tuple[Callable, Dict[str,str]]] = {}
    parameter_converter: Callable = staticmethod(lambda x: x)
    _context = {}
    _import_map = {"np": "numpy", "numpy":"numpy", "jnp":"jax.numpy", "jax.numpy": "jax.numpy"}

    def __init__(
        self, 
        name: str, 
        random_variable: RandomVariable, 
        dims: Tuple[str, ...],
        shape: Tuple[int, ...],
    ) -> None:
        self.name = name
        self._dist_str = random_variable.distribution
        self._parameter_expression = random_variable.parameters
        self._obs_transform: Optional[Expression] = random_variable.obs
        self._obs_transform_inv: Optional[Expression] = random_variable.obs_inv
        self.dims = dims
        self.shape = shape if len(shape) > 0 else ()

        dist, params, uargs = self.parse_distribution(random_variable)
        self.distribution: Callable = dist
        self.parameters: Dict[str, Expression] = params
        self.undefined_args: set = uargs

        self.obs_transform_func = self.create_transform_func(self._obs_transform, "obs_transform")
        self.obs_transform_func_inv = self.create_transform_func(self._obs_transform_inv, "obs_transform_inv")

    def __str__(self) -> str:
        dist = self.dist_name
        params = ", ".join([f"{k}={v}" for k, v in self.parameters.items()])
        dimshape = tuple([f'{d}={s}' for d, s in zip(self.dims, self.shape)])
        return f"{dist}({params}, dims={dimshape}, obs={self._obs_transform})"
    
    def __repr__(self) -> str:
        return str(self)

    @property
    def dist_name(self) -> str:
        return self._dist_str
    
    def create_transform_func(self, transform, func_name):
        if transform is None:
            return None
        
        else:
            expr = transform
            
            imports = [a for a in expr.undefined_args if a in self._import_map]
            args = [a for a in expr.undefined_args if a not in self._import_map]

            # Create the function arguments
            args = [ast.arg(arg=arg_name, annotation=None) for arg_name in args]
            
            # Build a function definition from the expression
            func_def = ast.FunctionDef(
                name=f"{func_name}_{self.name}",
                args=ast.arguments(
                    posonlyargs=[], args=args, vararg=None, kwonlyargs=[], 
                    kw_defaults=[], defaults=[]
                ),
                body=[ast.Return(value=expr.expression.body)],
                decorator_list=[]
            )
        
            import_statements = [
                ast.Import(names=[ast.alias(name=self._import_map[i], asname=i)])
                for i in imports
            ]
            module = ast.Module(body=[*import_statements, func_def], type_ignores=[])
            module = ast.fix_missing_locations(module)

            # compile the function and retrieve object
            code = compile(module, filename="<ast>", mode="exec")
            func_env = {}
            exec(code, func_env)    
            func = func_env[f"{func_name}_{self.name}"]
            
            return func

    
    def construct(self, context: Iterable[Mapping], extra_kwargs: Dict = {}):
        _context = {arg: lookup_from(arg, context) for arg in self.undefined_args}
        _context.update(self._context)
        # evaluate the parameters given a context
        params = {
            key: self.parameter_converter(value.evaluate(context=_context)) 
            for key, value in self.parameters.items()
        }
        return self.distribution(**params, **extra_kwargs)

    def _get_distribution(self, distribution: str) -> Tuple[Callable, Dict[str, str]]:
        return self.distribution_map[distribution]

    def parse_distribution(self, random_variable: RandomVariable) -> Tuple[Any,Dict[str,Expression],set]:

        distribution_mapping = self._get_distribution(random_variable.distribution)

        if not isinstance(distribution_mapping, tuple):
            distribution = distribution_mapping
            distribution_mapping = (distribution, {})
        
        assert len(distribution_mapping) == 2, (
            "distribution and parameter mapping must be "
            "a tuple of length 2."
        )

        distribution, parameter_mapping = distribution_mapping
        mapped_params = {}
        underfined_args = set()
        for key, val in random_variable.parameters.items():
            mapped_key = parameter_mapping.get(key, key)
            underfined_args = underfined_args.union(val.undefined_args)
            mapped_params.update({mapped_key:val})

        return distribution, mapped_params, underfined_args


class InferenceBackend(ABC):
    _distribution = Distribution
    idata: az.InferenceData
    prior: Dict[str,Distribution]
    log_likelihood: Errorfunction
    gradient_log_likelihood: Errorfunction
    chains = 1
    draws = 1

    def __init__(
        self, 
        simulation: SimulationBase,
    ) -> None:
        
        self.simulation = simulation
        self.config = simulation.config

        self.indices = {v.name: np.array(v.values) for _, v in self.simulation.indices.items()}
        # parse model components
        self.prior = self.parse_model_priors(
            parameters=self.config.model_parameters.free,
            dim_shapes=self.simulation.parameter_shapes,
            data_structure=self.config.data_structure,
            indices=self.indices
        )

        self.evaluator = self.parse_deterministic_model()

        self.error_model = self.parse_error_model(
            error_models=self.config.error_model.all
        )
        

    @abstractmethod
    def parse_deterministic_model(self):
        pass

    @abstractmethod
    def parse_probabilistic_model(self):
        pass

    @property
    def extra_vars(self):
        return self.config.inference.extra_vars
    
    @property
    def n_predictions(self):
        return self.config.inference.n_predictions
    
    @property
    def EPS(self):
        return self.config.inference.eps

    @property
    def posterior_data_structure(self) -> Dict[str, List[str]]:
        data_structure = self.simulation.data_structure.copy()
        data_structure_loglik = {f"{dv}_obs": dims for dv, dims in data_structure.items()}
        data_structure_residuals = {f"{dv}_res": dims for dv, dims in data_structure.items()}
        parameter_dims = {k: list(v) for k, v in self.simulation.parameter_dims.items() if len(v) > 0}
        parameter_dims_unconstrained = {
            f"{k}_normal_base": list(v) for k, v 
            in self.simulation.parameter_dims.items() if len(v) > 0
        }
        data_structure.update(data_structure_loglik)
        data_structure.update(data_structure_residuals)
        data_structure.update(parameter_dims)
        data_structure.update(parameter_dims_unconstrained)
        return data_structure
    
    @property
    def posterior_coordinates(self) -> Dict[str, List[str|int]]:
        chains = self.chains
        draws = self.draws

        posterior_coords = {k: list(v) for k, v in self.simulation.dimension_coords.items()}
        posterior_coords.update({
            "draw": list(range(draws)), 
            "chain": list(range(chains))
        })
        return posterior_coords

    @classmethod
    def parse_model_priors(
        cls, 
        parameters: Dict[str,Param], 
        dim_shapes: Dict[str,Tuple[int, ...]],
        data_structure: Datastructure,
        indices: Dict[str, Any] = {}
    ):
        priors = {}
        hyper_ = []
        for key, par in parameters.items():
            if par.prior is None:
                raise AttributeError(
                    f"No prior was defined for parameter '{key}'. E.g.: "+
                    f"`sim.config.model_parameters.{key}.prior = 'lognorm(loc=1, scale=2)'`"
                )
            
            for k, v in par.prior.parameters.items():
                for ua in v.undefined_args:
                    if ua in indices:
                        continue

                    elif ua in cls._distribution._import_map:
                        continue

                    elif ua in priors:
                        continue

                    elif ua in data_structure.data_variables:
                        # allows observations
                        continue

                    elif ua.strip("_y0") in data_structure.data_variables:
                        # allows initial values
                        continue

                    else:
                        raise KeyError(
                            f"Parameter '{key}' defines a prior '{par.prior.model_ser()}' that will try "+
                            f"to access the variable '{ua}' before it is defined "+
                            "in 'sim.indices', 'Distribution._import_map' or previously "+
                            "defined priors. Please double check the prior definition for errors, "+
                            f"specify the needed parameter, or check the parameter order. "+
                            "If needed, use "+
                            f"config.model_parameters.reorder([..., '{ua}', '{key}', ...]) "+
                            "to arrange the parameters in the correct order."
                        )

            dist = cls._distribution(
                name=key, 
                random_variable=par.prior,
                dims=par.dims,
                shape=dim_shapes[key]
            )
            priors.update({key: dist})
        return priors
    
    @classmethod
    def parse_error_model(
        cls,
        error_models: Dict[str,RandomVariable], 
    ):
        error_model = {}
        for data_var, error_distribution in error_models.items():
            error_dist = cls._distribution(
                name=data_var, 
                random_variable=error_distribution,
                dims=(),
                shape=(),
            )
               
            error_model.update({data_var: error_dist})
        return error_model

    @abstractmethod
    def create_log_likelihood(self) -> Tuple[Errorfunction,Errorfunction]:
        """This method creates a log likelihood function and potentially
        function to compute the gradients.
        """
        pass

    def plot_likelihood_landscape(
        self, 
        parameters: Tuple[str, str],
        log_likelihood_func: Callable,
        gradient_func: Optional[Callable] = None,
        bounds: Tuple[List[float],List[float]] = ([-10, 10], [-10,10]),
        n_grid_points: int = 100,
        n_vector_points: int = 50,
        ax: Optional[plt.Axes] = None,
    ):
        """Plots the likelihood for each coordinate pair of two model parameters
        Parameters are taken from the standardized scale and transformed 

        For some reason the likelihood and gradients need to be transposed after
        calculating to work with meshgrid. The coordinates can be 

        Parameters
        ----------

        parameters : Tuple[str,str]
            The parameters to be plotted against each other

        log_likelihood_func : Callable
            Must be a vectorized function that can dictionary with two keys that
            contain array values as input parameters and returns a 1-D array of 
            log-likelihood values.

        gradient_func : Callable
            If, in addition to the likelihood values, also gradients should
            be computed, a function to compute the gradients must be provided 

        bounds : Tuple[Tuple[float,float],Tuple[float,float]]
            Fallback bounds for the parameter grid to calculate the likelihood
            function. This is only used if bounds are not provided by the parameters

        n_grid_points : int
            The number of grid points per side, to calculate the likelihood
            function for. Scales n^2
        """
        par_x, par_y = parameters
        bounds_x, bounds_y = bounds
        px = self.config.model_parameters.all[par_x]
        py = self.config.model_parameters.all[par_y]

        bounds_x[0] = px.min if px.min is not None else bounds_x[0]
        bounds_x[1] = px.max if px.max is not None else bounds_x[1]
        bounds_y[0] = py.min if py.min is not None else bounds_y[0]
        bounds_y[1] = py.max if py.max is not None else bounds_y[1]

        x = np.linspace(*bounds_x, n_grid_points)
        y = np.linspace(*bounds_y, n_grid_points)

        grid = {p: v for p, v in zip(parameters, np.array(list(it.product(x, y))).T)}
        loglik = log_likelihood_func(grid)

        # loglikelihood must be transposed to work with meshgrid
        Z = loglik.reshape((n_grid_points, n_grid_points)).T
        X, Y = np.meshgrid(x, y)

        if ax is None:
            fig, ax = plt.subplots(1, 1, figsize=(8,6))

        cmap = mpl.colormaps["viridis"]  # type: ignore
        contours = ax.contourf(X, Y, Z, cmap=cmap, levels=50)

        if gradient_func is not None:
            xv = np.linspace(*bounds_x, n_vector_points)
            yv = np.linspace(*bounds_y, n_vector_points)
            Xv, Yv = np.meshgrid(xv, yv)

            gridv = {p: v for p, v in zip(parameters, np.array(list(it.product(xv, yv))).T)}

            grads = gradient_func(gridv)

            u, v = grads[par_x], grads[par_y]

            # gradients must be transposed to work with meshgrid
            U = u.reshape((n_vector_points, n_vector_points)).T
            V = v.reshape((n_vector_points, n_vector_points)).T

            vector_field = ax.quiver(Xv, Yv, U, V, angles="xy", width=0.001)

        ax.figure.colorbar(
            contours,
            ax=ax,
            label="log-likelihood"
        )
        ax.set_xlabel(f"{par_x} (normalized)")
        ax.set_ylabel(f"{par_y} (normalized)")

        return ax
    
    @abstractmethod
    def prior_predictions(self):
        pass

    @abstractmethod
    def posterior_predictions(self):
        pass

    def plot_prior_predictions(
            self, data_variable: str, x_dim: str, ax=None, subset={}, 
            n=None, seed=None, plot_preds_without_obs=False,
            prediction_data_variable: Optional[str] = None,
            **plot_kwargs
        ):
        if n is None:
            n = self.n_predictions

        if seed is None:
            seed = self.config.simulation.seed
        
        idata = self.prior_predictions(
            n=n, 
            # seed only controls the parameters samples drawn from posterior
            seed=seed
        )

        ax = self.plot_predictions(
            observations=self.simulation.observations,
            predictions=idata.prior_predictive, # type: ignore
            data_variable=data_variable,
            plot_preds_without_obs=plot_preds_without_obs,
            x_dim=x_dim,
            ax=ax,
            subset=subset,
            prediction_data_variable=prediction_data_variable,
            **plot_kwargs,
        )

        return ax

    def plot_posterior_predictions(
            self, data_variable: str, x_dim: str, ax=None, subset={},
            n=None, seed=None, plot_preds_without_obs=False,
            prediction_data_variable: Optional[str] = None,
            **plot_kwargs
        ):
        # TODO: This method should be trashed. It is not really useful
        if n is None:
            n = self.n_predictions
        
        if seed is None:
            seed = self.config.simulation.seed
        
        predictions = self.posterior_predictions(
            n=n, 
            # seed only controls the parameters samples drawn from posterior
            seed=seed
        )
        
        ax = self.plot_predictions(
            observations=self.simulation.observations,
            predictions=predictions,
            data_variable=data_variable,
            plot_preds_without_obs=plot_preds_without_obs,
            x_dim=x_dim,
            ax=ax,
            subset=subset,
            prediction_data_variable=prediction_data_variable,
            **plot_kwargs
        )

        return ax

    def plot(self):
        self.plot_diagnostics()

        plot = self.config.inference.plot
        if plot is None:
            return
        elif isinstance(plot, str):
            try:
                plot_func = getattr(self.simulation._plot, plot)
                plot_func(self.simulation)
            except AttributeError:
                warnings.warn(
                    f"Plot function {plot} was not found in the {self.simulation._plot} module "+
                    "Make sure the name has been spelled correctly or try to "+
                    "set the function directly to 'sim.config.inference.plot'.",
                    category=UserWarning
                )
        else:
            plot(self.simulation)

    @staticmethod
    def plot_predictions(
            observations,
            predictions,
            data_variable: str,
            x_dim: str, 
            ax=None, 
            plot_preds_without_obs=False,
            subset={},
            mode: Literal["mean+hdi", "draws"]="mean+hdi",
            plot_options: Dict={"obs": {}, "pred_mean": {}, "pred_draws": {}, "pred_hdi": {}},
            prediction_data_variable: Optional[str] = None,
        ):
        warnings.warn(
            "Use of 'sim.inferer.plot_predictions' is deprecated. Use the "+
            "Plotting backend: pymob.sim.plot.SimulationPlot",
            category=DeprecationWarning
        )
        # filter subset coordinates present in data_variable
        subset = {k: v for k, v in subset.items() if k in observations.coords}
        
        if prediction_data_variable is None:
            prediction_data_variable = data_variable

        # select subset
        if prediction_data_variable in predictions:
            preds = predictions.sel(subset)[prediction_data_variable]
        else:
            raise KeyError(
                f"{prediction_data_variable} was not found in the predictions "+
                "consider specifying the data variable for the predictions "+
                "explicitly with the option `prediction_data_variable`."
            )
        try:
            obs = observations.sel(subset)[data_variable]
        except KeyError:
            obs = preds.copy().mean(dim=("chain", "draw"))
            obs.values = np.full_like(obs.values, np.nan)
        
        # stack all dims that are not in the time dimension
        if len(obs.dims) == 1:
            # add a dummy batch dimension
            obs = obs.expand_dims("batch")
            obs = obs.assign_coords(batch=[0])

            preds = preds.expand_dims("batch")
            preds = preds.assign_coords(batch=[0])


        stack_dims = [d for d in obs.dims if d not in [x_dim, "chain", "draw"]]
        obs = obs.stack(i=stack_dims)
        preds = preds.stack(i=stack_dims)
        N = len(obs.coords["i"])
            
        hdi = az.hdi(preds, .95)[f"{prediction_data_variable}"]

        if ax is None:
            ax = plt.subplot(111)
        
        y_mean = preds.mean(dim=("chain", "draw"))

        for i in obs.i:
            if obs.sel(i=i).isnull().all() and not plot_preds_without_obs:
                # skip plotting combinations, where all values are NaN
                continue
            
            if mode == "mean+hdi":
                kwargs_hdi = dict(color="black", alpha=0.1)
                kwargs_hdi.update(plot_options.get("pred_hdi", {}))
                ax.fill_between(
                    preds[x_dim].values, *hdi.sel(i=i).values.T, # type: ignore
                    **kwargs_hdi
                )

                kwargs_mean = dict(color="black", lw=1, alpha=max(1/N, 0.05))
                kwargs_mean.update(plot_options.get("pred_mean", {}))
                ax.plot(
                    preds[x_dim].values, y_mean.sel(i=i).values, 
                    **kwargs_mean
                )
            elif mode == "draws":
                kwargs_draws = dict(color="black", lw=0.5, alpha=max(1/N, 0.05))
                kwargs_draws.update(plot_options.get("pred_draws", {}))
                ys = preds.sel(i=i).stack(sample=("chain", "draw"))
                ax.plot(
                    preds[x_dim].values, ys.values, 
                    **kwargs_draws
                )
            else:
                raise NotImplementedError(
                    f"Mode '{mode}' not implemented. "+
                    "Choose 'mean+hdi' or 'draws'."
                )

            kwargs_obs = dict(marker="o", ls="", ms=3, color="tab:blue")
            kwargs_obs.update(plot_options.get("obs", {}))
            ax.plot(
                obs[x_dim].values, obs.sel(i=i).values, 
                **kwargs_obs
            )
        
        ax.set_ylabel(data_variable)
        ax.set_xlabel(x_dim)

        return ax

    def plot_diagnostics(self):
        """
        TODO: Should be outsourced to pymob.sim.plot
        """
        if hasattr(self.idata, "posterior"):
            axes = az.plot_trace(
                self.idata,
                var_names=list(self.config.model_parameters.free.keys())
            )
            fig = plt.gcf()
            fig.tight_layout()
            fig.savefig(f"{self.simulation.output_path}/trace.png")
            axes = az.plot_pair(
                self.idata, 
                divergences=True, 
                var_names=list(self.config.model_parameters.free.keys())
            )
            fig = plt.gcf()
            fig.tight_layout()
            fig.savefig(f"{self.simulation.output_path}/pairs_posterior.png")

    def check_prior_for_nans(self, idata):
        nans = idata.prior.isnull().sum().to_array()
        if np.any(nans > 0):
            prior_names = nans.where(nans > 0, drop=True)["variable"].values
            warnings.warn(
                f"NaNs occurred in the prior draws of {list(prior_names)}. "+
                "Make sure your priors are correctly specified. "+
                "I.e. prior parameters are inside the support of the distribution "+
                "(negative values, out of bounds values), this is especially "+
                "relevant if hyper-priors are used. "+
                f"Prior: {self.prior}",
                category=UserWarning
            )
            return False
        else:
            return True


from typing import Iterable, Any
from typing_extensions import Protocol
import inspect
import json
import re
import sympy
import warnings
from configparser import ConfigParser

CONSTANTS = ConfigParser()
CONSTANTS.read("config/constants.cfg")

def read_config(config_file):
    with open(config_file, "r") as f:
        return json.load(f)

def simulation_io_adapter(input_config_file, input_events_file, output_path):
    """
    A simple adapter to adapt the configuration file to be usable with the
    case-study design for the example of experiment based simulation
    """
    
    cfg = read_config(config_file=input_config_file)

    cfg["experiment"]["eventfile"] = input_events_file
    cfg["simulation"]["output"] = output_path

    return cfg


def catch_patterns(expression_str):
    # AVOID USING THE DESING OF EXPRESSIONS. JUST USE SYMPY SYNTAX._FlagsType
    # THIS WILL BE MORE STABLE AND EASY TO ADAPT NEW CONCEPTS.
    # THE KEY IS NOT THE EXPRESSION BUT THE LOOKUP
    
    # tries to match array notation [0 1 2]
    pattern = r"\[(\d+(\.\d+)?(\s+\d+(\.\d+)?)*|\s*)\]"
    if re.fullmatch(pattern, expression_str) is not None:
        expression_str = expression_str.replace(" ", ",")
        return f"Array({expression_str})"

    # tries to match array notation [0,1,2]
    pattern = r'\[(\d+(\.\d+)?(\s*,\s*\d+(\.\d+)?)*|\s*)\]'
    if re.fullmatch(pattern, expression_str) is not None:
        return f"Array({expression_str})"

    return expression_str

def lambdify_expression(expression_str):
    # check for parentheses in expression
    
    expression_str = catch_patterns(expression_str)

    # Parse the expression without knowing the symbol names in advance
    parsed_expression = sympy.sympify(expression_str, evaluate=False)
    free_symbols = tuple(parsed_expression.free_symbols)

    # Transform expression to jax expression
    args = [str(s) for s in free_symbols]
    func = sympy.lambdify(
        args, parsed_expression
    )

    return func, args

from typing import Mapping

def lookup_from(val: Any, collection: Iterable[Mapping]) -> Any:
    if isinstance(collection, dict):
        collection = list(collection.values())

    for obj in collection:
        if val in obj:
            return obj[val]
        else:
            continue

    return val

def lookup(val, *indexable_objects):
    for obj in indexable_objects:
        if val in obj:
            return obj[val]
        else:
            continue

    return val

def lookup_args(args, *objects_to_search):
    return {k: lookup(k, *objects_to_search) for k in args}

def get_return_arguments(func):
    ode_model_source = inspect.getsource(func)
    
    # extracts last return statement of source
    return_statement = ode_model_source.split("\n")[-2]

    # extract arguments returned by ode_func
    return_args = return_statement.split("return")[1]

    # strip whitespace and separate by comma
    return_args = return_args.replace(" ", "").split(",")

    return return_args

def dX_dt2X(expr: str):
    expr = expr.split("_dt", 1)[0]
    expr = expr.split("_dx", 1)[0]

    if len(expr) == 2:
        if expr[0] == "d":
            expr = expr[1]
            return expr

    raise NotImplementedError(
        "Derviatives returned by an ODE model should follow the form "
        "'dX_dt' or 'dX_dx' or 'dX' where 'X' denotes the variable."
    )


import os
import sys
import click
import resource

from pymob.utils import help
from pymob.utils.store_file import prepare_casestudy, import_package
from pymob.simulation import SimulationBase
from pymob.sim.config import Config

@click.command()
@click.option("-c", "--case_study", type=str, default="lotka_volterra_case_study", 
              help=help.case_study)
@click.option("-s", "--scenario", type=str, default="test_scenario", 
              help=help.scenario)
@click.option("-p", "--package", type=str, default="case_studies", 
              help=help.package)
@click.option("-o", "--output", type=str, default=None, 
              help="Set `case-study.output` directory")
@click.option("-r", "--random_seed", type=int, default=None, 
              help="Set `simulation.seed`")
@click.option("-n", "--n_cores", type=int, default=None, 
              help="The number of cores to be used for multiprocessing")
@click.option("--inference_backend", type=str, default="pymoo")
def main(case_study, scenario, package, output, random_seed, n_cores, inference_backend):
    
    cfg = os.path.join(package, case_study, "scenarios", scenario, "settings.cfg")
    config = Config(cfg)
    config.case_study.name = case_study
    config.case_study.scenario = scenario
    config.case_study.package = package
    config.import_casestudy_modules()

    if n_cores is not None: config.multiprocessing.cores = n_cores
    if random_seed is not None: config.simulation.seed = random_seed
    if output is not None: config.case_study.output = output

    # import simulation      
    Simulation = config.import_simulation_from_case_study()
    sim = Simulation(config)
    sim.setup()
    sim.config.save(os.path.join(sim.output_path, "settings.cfg"), force=True)

    sim.set_inferer(backend=inference_backend)
    sim.prior_predictive_checks()
    sim.inferer.run()
    sim.inferer.store_results()
    sim.posterior_predictive_checks()
    sim.inferer.plot()

    max_ram_mb = resource.getrusage(resource.RUSAGE_SELF).ru_maxrss / 1000
    print("RESOURCE USAGE")
    print("==============")
    print(f"Max RSS: {max_ram_mb} M")


if __name__ == "__main__":
    main()


from typing import Callable
import sympy as sp
import numpy

from pymob.solvers.symbolic import FunctionPythonCode

def test_function_coder():
    code = FunctionPythonCode(
        expand_arguments=False, 
        lhs_0=(None, ()),
        theta=("theta", ("a", "b")),
        rhs=(sp.Symbol("b") * sp.Symbol("x") + sp.Symbol("a"),), # type: ignore
        lhs=("y",),
        x="x"
    )
    # this defines the function
    exec(
        f"{code}"
        "res = f(x=5, theta=(1,5))\n"
        "assert res == 26"
    )

    code.expand_arguments = True
    exec(
        f"{code}"
        "res = f(x=5, a=1, b=5)\n"
        "assert res == 26"
    )

def test_ode_coder():
    code = FunctionPythonCode(
        expand_arguments=False, 
        lhs_0=("Y_0", ("D_0", "H_0")),
        theta=("theta", ("b", "z")),
        rhs=(
            sp.Symbol("D_0") * sp.Symbol("t") + sp.Symbol("b"), # type: ignore
            sp.Symbol("H_0") * sp.Symbol("t") - sp.Symbol("z"), # type: ignore
        ), 
        lhs=("D", "H"),
        x="t"
    )

    # import numpy, needs to happen at the script level
    
    exec(
        f"{code}"
        "res = f(t=2, Y_0=(2,5), theta=(1,2))\n"
        "assert res == (2 * 2 + 1, 5 * 2 - 2)"
    )

    code.expand_arguments = True
    exec(
        f"{code}"
        "res = f(t=2, D_0=2, H_0=5, b=1, z=2)\n"
        "assert res == (2 * 2 + 1, 5 * 2 - 2)"
    )


def test_dummy_function():
    code = FunctionPythonCode(expand_arguments=True, lhs=(), rhs=())
    res = exec(str(code))
    assert res is None


if __name__ == "__main__":
    test_ode_coder()
    pass


from typing import Callable, Dict, List, Optional, Union, List
import xarray as xr

def stack_variables(
    ds: xr.Dataset, 
    variables: List[str], 
    new_coordinates: List[str],
    new_dim: str,
    pattern: Callable = lambda var, coord: f"{var}_{coord}"
):
    """Combine data variables and coordinates into a new variable of 
    a higher dimension.

    Parameters
    ----------
    ds : xr.Dataset
        The input xarray Dataset.
    variables : List[str]
        List of variable bases to stack.
    new_coordinates : List[str]
        List of new coordinates for the higher dimension. Note that this must
        be the same for all variables you want to stack.
    new_dim : str
        The name of the new dimension.
    pattern : Callable, optional
        A function to generate names for new variables based on the variable and coordinate names.

    Returns
    -------
    xr.Dataset
        The modified xarray Dataset with stacked variables.


    Examples
    --------     
    A use case is the following. Consider a dataset that has the variables

        cext_A
        cext_B
        cext_C
        my_other_var_1
        my_other_var_2
        
    And you want to combine the variables with the same base `cext_` into
    a new variable that has the base as a dimension.

    >>> import xarray as xr
    >>> from typing import List, Callable
    >>> from pymob.sim.evaluators import stack_variables
    >>> 
    >>> # Example usage:
    >>> an_xarray_dataset = xr.Dataset({
    ...     'cext_A': ([], 1.0),
    ...     'cext_B': ([], 2.0),
    ...     'cext_C': ([], 3.0),
    ...     'my_other_var_1': ([], 4.0),
    ...     'my_other_var_2': ([], 5.0),
    ... })
    >>> 
    >>> result_dataset = stack_variables(
    ...     ds=an_xarray_dataset,
    ...     variables=["cext"],
    ...     new_coordinates=["A", "B", "C"],
    ...     new_dim="letters",
    ... )
    >>> 
    >>> print(result_dataset)
    <xarray.Dataset>
    Dimensions:  (letters: 3)
    Coordinates:
      * letters  (letters) <U1 'A' 'B' 'C'
    Data variables:
        my_other_var_1  (float64) float64 4.0
        my_other_var_2  (float64) float64 5.0
        cext     (letters) float64 1.0 2.0 3.0
    """
    for v in variables:
        v_names = [pattern(v, c) for c in new_coordinates]
        
        # convert coordinate if existing to a data_variable
        v_in_coords = False
        if all([vn in ds.coords for vn in v_names]):
            ds = ds.reset_coords(v_names)
            v_in_coords = True
        
        var: xr.Dataset = ds[v_names]
        array = var.to_array(dim=new_dim).assign_coords({new_dim:new_coordinates})
        ds[v] = array

        ds = ds.drop_vars(v_names)
        
        if v_in_coords:
            ds = ds.set_coords(v)

    return ds


def unlist_attrs(ds: xr.Dataset|xr.DataArray):
    """Transforms lists of variables to a comma separated string to 
    work around errors when storing the dataset or dataarray to disk"""
    for key, value in ds.attrs.items():
        if isinstance(value, list):
            ds.attrs[key] = ", ".join(value)

    return ds


def enlist_attr(ds: xr.Dataset|xr.DataArray, attr: str):
    """Transforms a string representation of a metadata attribute of an
    xarray dataset or datarray to a list"""
    attr_string = ds.attrs[attr]
    assert isinstance(attr_string, str)
    
    ds.attrs[attr] = attr_string.split(", ")

    return ds



import numpy as np
import pytest
from tests.fixtures import (
    init_simulation_casestudy_api,
    init_guts_casestudy_constant_exposure,
    init_guts_casestudy_variable_exposure,
    init_bufferguts_casestudy,
)
import jax
import jax.numpy as jnp

# this should raise an exception when nans are created the computation


def test_guts_constant_exposure():
    sim = init_guts_casestudy_constant_exposure()

    sim.use_jax_solver()
    sim.dispatch_constructor(rtol=1e-7, atol=1e-8)

    # test gradients
    def test_func(theta):
        k_d, h_b, b, z = theta
        e = sim.dispatch({"k_d": k_d, "h_b": h_b, "b":b, "z": z})
        e()
        return e.Y["S"][-1]

    val, grads = jax.value_and_grad(test_func)(jnp.array([1, 0.001, 0.2, 1.0]))

    assert all(grads != jnp.nan)

def test_guts_variable_exposure():
    sim = init_guts_casestudy_variable_exposure()

    sim.use_jax_solver()
    sim.dispatch_constructor(rtol=1e-7, atol=1e-8)

    # test gradients
    def test_func(theta):
        k_d, h_b, b, z = theta
        e = sim.dispatch({"k_d": k_d, "h_b": h_b, "b":b, "z": z})
        e()
        return e.Y["S"][-1]

    val, grads = jax.value_and_grad(test_func)(jnp.array([1, 0.001, 0.2, 1.0]))

    assert all(grads != jnp.nan)


def test_bufferguts_variable_exposure():
    # Can be fixed by replacing the bufferGUTS model with the improved
    # version from the jupyter notebook
    sim = init_bufferguts_casestudy()

    sim.use_jax_solver()
    sim.dispatch_constructor(rtol=1e-7, atol=1e-8, max_steps=1e6)

    # test gradients
    def test_func(theta):
        k_d, h_b, b, z = theta
        e = sim.dispatch({"k_d": k_d, "h_b": h_b, "b":b, "z": z})
        e()
        return e.Y["S"][-1]

    val, grads = jax.value_and_grad(test_func)(jnp.array([1, 0.001, 0.2, 1.0]))

    assert all(grads != jnp.nan)

if __name__ == "__main__":
    import sys
    import os
    sys.path.append(os.getcwd())
    pass
    # test_multiple_case_study_import()
    # test_variable_exposure()
    # test_bufferguts_variable_exposure()
    # test_bufferguts_hybrid_solution()



from typing import Literal, Dict, Optional, List, Callable, TypeAlias
import warnings

import xarray as xr
import arviz as az
import numpy as np
import numpy.typing as npt

from pymob.sim.config import Config
from matplotlib import pyplot as plt
from matplotlib.figure import Figure
from matplotlib.axes import Axes

ALLOWED_IDATA_GROUPS: TypeAlias = Literal[
    "prior_predictive", 
    "prior_model_fits",
    "posterior_predictive", 
    "posterior_model_fits"
]


class SimulationPlot:
    """
    Parameters
    ----------

    observations : xr.Dataset
        Simulation observations dataset 

    idata : az.InferenceData
        Arviz InferenceData created from prior or posterior predictions.

    coordinates : Dict
        Coordinates of the simulation

    config : Config
        Simulation configuration

    rows : List[str]
        Optional. List of variables to plot in the rows of the figure. Defaults
        to the names of the data variables

    columns : str|Dict[str,List[str|int]]
        Optional. Group identifiers. Can be a single string (in this case 
        should be a dimension of the data structure). Can be a dictionary that
        has a single key (the dimension) and a list of values (a subset of)
        the dimension coordinates.
        
    idata_groups : List[str]
        Optional. The idata groups to plot for predictions. Only works for groups
        that have the same data structure as the observations. These are
        
    obs_idata_map : Dict[str|str|callable]
        Optional. Maps the data variables of the observation to the data variables
        of the inference data. This helps in cases, where the data variable of 
        the idata do not have the same name as the data variable of the observations.
        It can also be a callable that extracts the data variable from the 
        idataset.
    """
    figure: Figure
    axes_map: Dict[str,Dict[str,Axes]]
    def __init__(
        self, 
        observations: xr.Dataset,
        idata: az.InferenceData,
        coordinates: Dict,
        config: Config,
        rows: Optional[List[str]] = None,
        columns: Optional[str|Dict[str,List[str|int]]] = None,
        obs_idata_map: Dict[str,str|Callable] = {},
        idata_groups: Optional[ALLOWED_IDATA_GROUPS] = None,
        pred_mode: str = "mean+hdi",
        hdi_interval: float = 0.95,
        obs_style: Dict = {},
        pred_mean_style: Dict = {},
        pred_hdi_style: Dict = {},
        pred_draws_style: Dict = {},
        sharex=True,
        sharey="row",
    ):
        self.observations = observations
        self.idata = idata
        self.coordinates: Dict = coordinates
        self.config: Config = config

        if rows is None:
            self.rows = self.config.data_structure.observed_data_variables
        else:
            self.rows = rows

        if columns is None:
            self.columns = ["all"]
        else:
            self.columns = columns

        self.idata_groups = idata_groups

        self.obs_idata_map = obs_idata_map

        self.sharex = sharex
        self.sharey = sharey

        self.obs_style = dict(marker="o", ls="", ms=3, color="tab:blue")
        self.obs_style.update(obs_style)

        self.pred_mode = pred_mode
        self.hdi_interval = hdi_interval
        
        batch_idx = self.coordinates.get(self.config.simulation.batch_dimension, None)
        if batch_idx is None:
            N = 1
        else:
            N = len(batch_idx)

        self.pred_mean_style = dict(color="black", alpha=0.1)
        self.pred_mean_style.update(pred_mean_style)

        self.pred_hdi_style = dict(color="black", lw=1, alpha=max(1/N, 0.05))
        self.pred_hdi_style.update(pred_hdi_style)

        self.pred_draws_style = dict(color="black", lw=1, alpha=max(1/N, 0.05))
        self.pred_draws_style.update(pred_draws_style)
        
        self.create_figure()
        self.inf_preds: Dict[str: xr.DataArray] = {}


    def create_figure(self):
        r = len(self.rows)
        c = len(self.columns)
        self.figure, axes = plt.subplots(
            nrows=r, ncols=c,
            sharex=self.sharex,
            sharey=self.sharey,
            squeeze=False,
            figsize=(5+(c-1*2), 3+(r-1)*2),
        )

        self.axes_map = {}
        for i, row in enumerate(self.rows):
            self.axes_map[row] = {}
            for j, col in enumerate(self.columns):
                self.axes_map[row][col] = axes[i,j]

    def clean_idata_group(self, idata_group: str):
        # this mask is rigid. It will eliminate each draw where any of the 
        # variables contained an inf value
        group = self.idata[idata_group]
        
        stack_dims = [d for d in group.dims if d not in ["chain", "draw"]]

        mask = (group == np.inf).sum(stack_dims).to_array().sum("variable") > 0
        n_inf = int(mask.sum())
        if n_inf > 0:
            warnings.warn(
                f"There were {n_inf} NaN or Inf values in the idata group "+
                f"'{idata_group}'. See "+
                "Simulation.inf_preds for a mask with the coordinates.",
                category=UserWarning
            )

        self.inf_preds.update({idata_group: mask})

        return group.where(~mask, drop=True)

    def plot_data_variables(self):
        for i, row in enumerate(self.rows):
            for j, col in enumerate(self.columns):
                ax = self.axes_map[row][col]
                for igroup in self.idata_groups:
                    self.plot_predictions(idata_group=igroup, data_variable=row, column=col, ax=ax)

                if self.config.data_structure.all[row].observed:
                    self.plot_observations(data_variable=row, column=col, ax=ax)

        self.figure.tight_layout()
        
    def plot_observations(
        self,
        data_variable: str,
        column: str,
        ax: Axes,
    ):
        x_dim = self.config.simulation.x_dimension
        observations = self.observations[data_variable].copy()
        # stack all dims that are not in the time dimension
        if len(observations.dims) == 1:
            # add a dummy batch dimension
            observations = observations.expand_dims("batch")
            observations = observations.assign_coords(batch=[0])


        stack_dims = [d for d in observations.dims if d not in [x_dim]]
        observations = observations.stack(i=stack_dims)
        N = len(observations.coords["i"])
            
        for i in observations.i:
            obs = observations.sel(i=i)
            if obs.isnull().all():
                # skip plotting combinations, where all values are NaN
                continue
            
            self.plot_single_observation(obs, ax)

        ax.set_xlabel(x_dim)
    
    def plot_predictions(
        self,
        idata_group: str,
        data_variable: str,
        column: str,
        ax: Axes, 
    ):
        x_dim = self.config.simulation.x_dimension
        idata_dims = ["chain", "draw"]

        # get idata dataset and remove inf draws from the prior/posterior
        idata_dataset = self.clean_idata_group(idata_group)
        idata_dataset.attrs.update({"group": idata_group})

        prediction_data_variable = self.obs_idata_map.get(
            data_variable, data_variable
        )

        # this passes the idata group to the callable in the obs_idata_map
        # and returns it. This is very helpful if the prediction data variable
        # needs to be transformed prior to plotting 
        if callable(prediction_data_variable):
            predictions = prediction_data_variable(idata_dataset)
            prediction_data_variable = predictions.name
        else:
            if prediction_data_variable in idata_dataset:
                predictions = idata_dataset[prediction_data_variable].copy()
            else:
                return

        # stack all dims that are not in the time dimension
        if len([d for d in predictions.dims if d not in idata_dims]) == 1:
            # add a dummy batch dimension
            predictions = predictions.expand_dims("batch")
            predictions = predictions.assign_coords(batch=[0])


        stack_dims = [d for d in predictions.dims if d not in [x_dim] + idata_dims]
        predictions = predictions.stack(i=stack_dims)
        N = len(predictions.coords["i"])
            


        for i in predictions.i:
            # if obs.sel(i=i).isnull().all() and not plot_preds_without_obs:
            #     # skip plotting combinations, where all values are NaN
            #     continue
            preds = predictions.sel(i=i)

            self.plot_single_prediction(predictions=preds, ax=ax)
            
        
        ax.set_ylabel(data_variable)
        ax.set_xlabel(x_dim)


    def plot_single_prediction(self, predictions: xr.DataArray, ax: Axes):

        modes = self.pred_mode.split("+")

        for mode in modes:
            if mode == "hdi":
                self.plot_pred_hdi(predictions, ax)

            elif mode == "mean":
                self.plot_pred_mean(predictions, ax)

            elif mode == "draws":
                self.plot_pred_draws(predictions, ax)

            else:
                raise NotImplementedError(
                    f"Mode '{self.pred_mode}' not implemented. "+
                    "Choose 'mean', 'hdi', 'draws' or any combination of them, "+
                    "e.g., 'hdi+mean' (in this case the order determines the "+
                    "order of plots)"
                )

    def plot_single_observation(self, observations: xr.DataArray, ax: Axes):
        x_dim = self.config.simulation.x_dimension
        ax.plot(
            observations[x_dim].values, observations.values, 
            **self.obs_style
        )

    def plot_pred_hdi(self, predictions: xr.DataArray, ax: Axes):
        x_dim = self.config.simulation.x_dimension
        hdi = az.hdi(predictions, self.hdi_interval)[predictions.name]
        ax.fill_between(
            predictions[x_dim].values, *hdi.values.T, # type: ignore
            **self.pred_hdi_style
        )
    
    def plot_pred_mean(self, predictions: xr.DataArray, ax: Axes):
        x_dim = self.config.simulation.x_dimension
        y_mean = predictions.mean(dim=("chain", "draw"))
        ax.plot(
            predictions[x_dim].values, y_mean.values, 
            **self.pred_mean_style
        )

    def plot_pred_draws(self, predictions: xr.DataArray, ax: Axes):
        x_dim = self.config.simulation.x_dimension
        ys = predictions.stack(sample=("chain", "draw"))
        ax.plot(
            predictions[x_dim].values, ys.values, 
            **self.pred_draws_style
        )
    
    def set_titles(self, title: Callable):
        for c in self.columns:
            ax = self.axes_map[self.rows[0]][c]
            ax.set_title(label=title(self, c))

        self.figure.tight_layout()

    def close(self):
        plt.close(self.figure)

    def save(self, filename):
        self.figure.savefig(
            f"{self.config.case_study.output_path}/{filename}"
        )

        


import pytest
from matplotlib import pyplot as plt
from pymob.solvers import JaxSolver
import numpy as np
import jax

from tests.fixtures import (
    init_simulation_casestudy_api,
    init_lotka_volterra_case_study_hierarchical_from_script,
    init_lotka_volterra_case_study_hierarchical_from_settings
)


def test_inference_evaluation():
    pytest.skip()
    sim = init_simulation_casestudy_api()
    sim.set_inferer(backend="pyabc")

    sim.inferer.load_results()
    fig = sim.inferer.plot_chains()
    fig.savefig(sim.output_path + "/pyabc_chains.png")

    # posterior predictions
    for data_var in sim.config.data_structure.data_variables:
        ax = sim.inferer.plot_predictions(
            data_variable=data_var, 
            prediction_data_variable=data_var,
            x_dim="time"
        )
        fig = ax.get_figure()

        fig.savefig(f"{sim.output_path}/pyabc_posterior_predictions_{data_var}.png")
        plt.close()


def test_prior_predictions():
    sim = init_lotka_volterra_case_study_hierarchical_from_settings("lotka_volterra_hierarchical_presimulated_v1")
    sim.solver = JaxSolver

    sim.config.inference.n_predictions = 30
    sim.config.inference_numpyro.chains = 2
    sim.config.inference_numpyro.draws = 5
    sim.config.inference_numpyro.nuts_max_tree_depth = 1
    sim.config.inference_numpyro.gaussian_base_distribution = True

    sim.dispatch_constructor()
    sim.set_inferer(backend="numpyro")

    idata = sim.inferer.prior_predictions(n=5)

    np.testing.assert_equal(
        list(idata.prior.data_vars.keys()),
        ["alpha", "alpha_sigma", "alpha_species", "alpha_species_hyper", "beta"]
    )
    
    np.testing.assert_equal(
        list(idata.prior_predictive.data_vars.keys()),
        ["rabbits", "wolves"]
    )

    np.testing.assert_equal(
        list(idata.prior_model_fits.data_vars.keys()),
        ["rabbits", "wolves"]
    )

    np.testing.assert_equal(
        list(idata.prior_residuals.data_vars.keys()),
        ["rabbits", "wolves"]
    )

    np.testing.assert_array_equal(idata.prior.coords["id"], sim.observations["id"])
    np.testing.assert_array_equal(idata.prior.coords["rabbit_species"], sim.dimension_coords["rabbit_species"])
    np.testing.assert_array_equal(idata.prior.coords["experiment"], sim.dimension_coords["experiment"])



def test_prior_predictions_nuts():
    sim = init_lotka_volterra_case_study_hierarchical_from_settings("lotka_volterra_hierarchical_presimulated_v1")
    sim.solver = JaxSolver

    sim.config.inference.n_predictions = 30
    sim.config.inference_numpyro.chains = 2
    sim.config.inference_numpyro.draws = 5
    sim.config.inference_numpyro.nuts_max_tree_depth = 1
    sim.config.inference_numpyro.gaussian_base_distribution = True
    sim.config.inference_numpyro.kernel = "nuts"

    sim.dispatch_constructor()
    sim.set_inferer(backend="numpyro")

    idata = sim.inferer.prior_predictions(n=5)

    np.testing.assert_equal(
        list(idata.prior.data_vars.keys()),
        ["alpha", "alpha_sigma", "alpha_species", "alpha_species_hyper", "beta"]
    )
    
    np.testing.assert_equal(
        list(idata.prior_predictive.data_vars.keys()),
        ["rabbits", "wolves"]
    )

    np.testing.assert_equal(
        list(idata.prior_model_fits.data_vars.keys()),
        ["rabbits", "wolves"]
    )

    np.testing.assert_equal(
        list(idata.prior_residuals.data_vars.keys()),
        ["rabbits", "wolves"]
    )

    np.testing.assert_array_equal(idata.prior.coords["id"], sim.observations["id"])
    np.testing.assert_array_equal(idata.prior.coords["rabbit_species"], sim.dimension_coords["rabbit_species"])
    np.testing.assert_array_equal(idata.prior.coords["experiment"], sim.dimension_coords["experiment"])


def test_posterior_predictions_nuts():
    sim = init_simulation_casestudy_api("test_scenario")
    sim.solver = JaxSolver

    sim.config.inference.n_predictions = 30
    sim.config.inference_numpyro.chains = 2
    sim.config.inference_numpyro.draws = 5
    sim.config.inference_numpyro.warmup = 0
    sim.config.inference_numpyro.nuts_max_tree_depth = 1
    sim.config.inference_numpyro.gaussian_base_distribution = True
    sim.config.inference_numpyro.kernel = "nuts"

    sim.dispatch_constructor()
    sim.set_inferer(backend="numpyro")

    sim.inferer.run()
    idata = sim.inferer.idata

    np.testing.assert_equal(
        list(idata.posterior.data_vars.keys()),
        ["alpha", "beta"]
    )
    
    np.testing.assert_equal(
        list(idata.posterior_predictive.data_vars.keys()),
        ["rabbits", "wolves"]
    )

    np.testing.assert_equal(
        list(idata.posterior_model_fits.data_vars.keys()),
        ["rabbits", "wolves"]
    )

    np.testing.assert_equal(
        list(idata.unconstrained_posterior.data_vars.keys()),
        ["alpha_normal_base", "beta_normal_base"]
    )


def test_posterior_predictions_svi():
    sim = init_simulation_casestudy_api("test_scenario")
    sim.solver = JaxSolver

    sim.config.inference_numpyro.svi_iterations = 10
    sim.config.inference_numpyro.draws = 5
    sim.config.inference_numpyro.gaussian_base_distribution = True

    sim.dispatch_constructor()
    sim.set_inferer(backend="numpyro")

    sim.config.inference_numpyro.kernel = "svi"
    sim.inferer.run()
    idata = sim.inferer.idata

    np.testing.assert_equal(
        list(idata.posterior.data_vars.keys()),
        ["alpha", "beta"]
    )
    
    np.testing.assert_equal(
        list(idata.posterior_predictive.data_vars.keys()),
        ["rabbits", "wolves"]
    )

    np.testing.assert_equal(
        list(idata.posterior_model_fits.data_vars.keys()),
        ["rabbits", "wolves"]
    )

    np.testing.assert_equal(
        list(idata.unconstrained_posterior.data_vars.keys()),
        ["alpha_normal_base", "beta_normal_base"]
    )

def test_model_check():
    sim = init_lotka_volterra_case_study_hierarchical_from_settings("lotka_volterra_hierarchical_presimulated_v1")
    sim = init_simulation_casestudy_api()
    sim.config.inference_numpyro.gaussian_base_distribution = True
    sim.config.jaxsolver.throw_exception = False
    sim.config.jaxsolver.max_steps = 1000

    sim.solver = JaxSolver
    sim.dispatch_constructor()
    sim.set_inferer(backend="numpyro")

    sim.inferer.check_gradients()
    sim.inferer.check_log_likelihood()


def test_vector_field():
    sim = init_lotka_volterra_case_study_hierarchical_from_settings("lotka_volterra_hierarchical_presimulated_v1")
    sim = init_simulation_casestudy_api()
    sim.config.inference_numpyro.gaussian_base_distribution = True
    sim.config.jaxsolver.throw_exception = False
    sim.config.jaxsolver.max_steps = 10_000

    sim.solver = JaxSolver
    sim.dispatch_constructor()
    sim.set_inferer(backend="numpyro")

    sim.config.model_parameters.beta.min = -10
    sim.config.model_parameters.beta.max = 10
    sim.config.model_parameters.alpha.min = -10
    sim.config.model_parameters.alpha.max = 10

    # Define a scalar function of two variables
    def f(theta):
        x = theta["alpha"]
        y = theta["beta"]
        return -((x+2)**2 + (y-2)**2)

    # Compute the gradient function
    gradient_f = jax.grad(f)

    ax = sim.inferer.plot_likelihood_landscape(
        parameters=("alpha", "beta"),
        log_likelihood_func=jax.vmap(f),
        gradient_func=jax.vmap(gradient_f),
    )
    ax.plot(-2,2,ls="", marker="o", color="black")
    ax.figure.savefig(f"{sim.output_path}/test_loglikelihood_gradients.png")

    log_likelihood, grad_log_likelihood = sim.inferer.create_log_likelihood(
        seed=1, return_type="custom", check=False, 
        custom_return_fn=lambda lj, lp, ld: lj,
        vectorize=True,
        gradients=True
    )


    ax = sim.inferer.plot_likelihood_landscape(
        parameters=("beta", "alpha"),
        log_likelihood_func=log_likelihood,
        gradient_func=grad_log_likelihood,
    )

    ax.figure.savefig(f"{sim.output_path}/loglikelihood_gradients.png")


if __name__ == "__main__":
    import sys
    import os
    sys.path.append(os.getcwd())
    # test_scripting_api_pyabc()


import pytest
import json
import numpy as np
from click.testing import CliRunner

from tests.fixtures import init_simulation_casestudy_api


def test_convergence():
    sim = init_simulation_casestudy_api()
    sim.set_inferer(backend="pymoo")
    sim.inferer.run()

    with open(f"{sim.config.case_study.output_path}/pymoo_params.json", "r") as f:
        pymoo_results = json.load(f)

    estimated_parameters = pymoo_results["X"]
    true_parameters = sim.model_parameter_dict
    
    np.testing.assert_allclose(
        np.array(list(estimated_parameters.values())),
        np.array(list(true_parameters.values())),
        rtol=1e-1, atol=5e-3
    )


def test_commandline_api_infer():
    # TODO: This will run, once methods are available for 
    # - prior_predictive_checks, 
    # - store_results, 
    # - posterior_predictive_checks 
    pytest.skip()
    from pymob.infer import main
    runner = CliRunner()
    
    args = "--case_study=lotka_volterra_case_study "\
        "--scenario=test_scenario "\
        "--inference_backend=pymoo"
    result = runner.invoke(main, args.split(" "))

    if result.exception is not None:
        raise result.exception


if __name__ == "__main__":
    import sys
    import os
    sys.path.append(os.getcwd())
    # test_scripting_api_pyabc()